#include <stdio.h>
#include <ctype.h>
#include "z80.h"
#include "zacvt.h"
#include "console.h"

extern unsigned char mem[MSIZE];
extern z80_processor z80;
extern int reroute;

void console::print1(const char *s)
{
	printf("%s: %d %X: %d %X %c\n", s, ptr, ptr, mem[ptr], mem[ptr],
	       mem[ptr]<0x40 ? zx2ascii[mem[ptr]] : '_');
	ptr++;
}

void console::print2(const char *s)
{
	printf("%s: %d %X: %d %X\n", s, ptr, ptr, mem[ptr]+(mem[ptr+1]<<8), mem[ptr]+(mem[ptr+1]<<8));
	ptr += 2;
}

void console::system_variables()
{
	ptr = 0x4000;
	print1("ERRNR  ");
	print1("FLAGS  ");
	print2("ERR-SP ");
	print2("RAMTOP ");
	print1("MODE   ");
	print2("PPC    ");
	print1("VERSN  ");
	print2("E_PPC  ");
	print2("D_FILE ");
	print2("DF-CC  ");
	print2("VARS   ");
	print2("DEST   ");
	print2("E-LINE ");
	print2("CH-ADD ");
	print2("X-PTR  ");
	print2("STKBOT ");
	print2("STKEND ");
	print1("BERG   ");
	print2("MEM    ");
	print1("N/U    ");
	print1("DF-SZ  ");
	print2("S-TOP  ");
	print2("LAST-K ");
	print1("DEBNCE ");
	print1("MARGIN ");
	print2("NXTLIN ");
	print2("OLDPPC ");
	print1("FLAGX  ");
	print2("STRLEN ");
	print2("T-ADDR ");
	print2("SEED   ");
	print2("FRAMES ");
	print2("COORDS ");
	print1("PR-CC  ");
	print1("S-POSN ");
	print1("S-POSN ");
	print1("CDFLAG ");
}

void console::vlist()
{
	unsigned char c, mask;

	z80.h = mem[0x4011]; /* vars */
	z80.l = mem[0x4010];
	while (1) {
	  ptr = (z80.h << 8) + z80.l;
	  c = mem[ptr];
	  if (c==0x80) break;
	  mask = c & 0xe0;
	  switch (mask) {
	    case 0x60 : printf("%c\n", zx2ascii[c&0x3f]);
			break;
	    case 0xa0 : printf("%c", zx2ascii[c&0x3f]);
			while (1) {
		          c = mem[++ptr];
			  printf("%c", zx2ascii[c&0x3f]);
			  if ((c&0xc0)==0x80) { printf("\n"); break; }
			}
			break;
	    case 0x40 : printf("%c$ [length: %d]\n", zx2ascii[(c&0x3f)+0x20], mem[ptr+1]);
			break;
	    case 0x80 : printf("%c  [dimensions: %d]\n", zx2ascii[(c&0x3f)+0x20], mem[ptr+3]);
			break;
	    case 0xc0 : printf("%c$ [dimensions: %d]\n", zx2ascii[(c&0x3f)+0x20], mem[ptr+3]);
			break;
	    case 0xe0 : printf("%c  [control variable]\n", zx2ascii[c&0x3f]);
			break;
	    default   :;
	  }
	  z80.run_ret(0x09f2);
	  z80.h = z80.d;
	  z80.l = z80.e;
	}
}

void console::save_vars()
{
	FILE *out;
	unsigned int ptr, n;

	out = fopen("vars.bin","wb");
	if (!out) {
	  fprintf(stderr, "zce: error saving variables to `vars.bin`\n");
	  return;
	}

	ptr = (mem[0x4011] << 8) | mem[0x4010];
	n = ((mem[0x4015] << 8) | mem[0x4014]) - ptr;

	fwrite(mem+ptr,1,n,out);
	fclose(out);
}

void console::merge_vars()
{
	FILE *in;
	unsigned int ptr, sptr, wptr, n;
	unsigned char buffer[0x10000];
	unsigned char c, vr, mask;

	in = fopen("vars.bin","rb");
	if (!in) {
	  fprintf(stderr, "zce: error loading variables from `vars.bin`\n");
	  return;
	}

	fread(buffer,1,0x10000,in);
	fclose(in);

	ptr = 0;
	while (1) {
	  sptr = ptr;
	  c = buffer[ptr];
	  if (c==0x80) break;
	  mask = c & 0xe0;
	  if (mask&0x30) vr = c&0x3f; else vr = (c&0x3f) + 0x20;

	  z80.b=0; z80.c=1; z80.run_ret(0x0030);
	  wptr = (z80.d << 8) | z80.e;
	  mem[wptr] = vr;
//	  printf("%c\n",zx2ascii[vr]);
	  switch (mask) {
	    case 0x60 : ptr += 6;
			break;
	    case 0xa0 : while (1) {
		          c = buffer[++ptr];
			  vr = c&0x3f;
	  		  z80.b=0; z80.c=1; z80.run_ret(0x0030);
			  mem[(z80.d << 8) | z80.e] = vr;
			  if ((c&0xc0)==0x80) break;
			}
			ptr += 6;
			break;
	    case 0x40 : z80.b=0; z80.c=1; z80.run_ret(0x0030);
			n = (buffer[ptr+2] << 8) | buffer[ptr+1];
			mem[(z80.d << 8) | z80.e] = ascii2zx['$'];
			ptr += 2+n+1;
			break;
	    case 0x80 : z80.b=0; z80.c=1; z80.run_ret(0x0030);
			mem[(z80.d << 8) | z80.e] = ascii2zx['('];
			n = (buffer[ptr+2] << 8) | buffer[ptr+1];
			ptr += 2+n+1;
			break;
	    case 0xc0 : z80.b=0; z80.c=1; z80.run_ret(0x0030);
			mem[(z80.d << 8) | z80.e] = ascii2zx['$'];
			n = (buffer[ptr+2] << 8) | buffer[ptr+1];
			ptr += 2+n+1;
			break;
	    case 0xe0 : ptr += 18;
			break;
	    default   : return;
	  }

	  z80.b=0; z80.c=1; z80.run_ret(0x0030);
	  mem[(z80.d << 8) | z80.e] = 0x76;

	  mem[0x4016] = wptr & 0x00ff; /* ch-add */
	  mem[0x4017] = (wptr & 0xff00) >> 8;

	  z80.run_ret(0x111c); /* look-vars */

	  if (z80.f&0x01) { /* if new variable, add to vars */
	    wptr = (mem[0x4015] << 8) | mem[0x4014]; /* e-line */
	    wptr--; /* 80 byte */
	    z80.h = (wptr & 0xff00) >> 8;
	    z80.l = wptr & 0x00ff;
	    n = ptr - sptr;
	    z80.b = (n & 0xff00) >> 8;
	    z80.c = n & 0x00ff;
	    z80.run_ret(0x99e); /* make-room */
	    do {
	      mem[wptr++] = buffer[sptr++];
	    } while (sptr<ptr);
	  }

	  z80.run_ret(0x14a3); /* x-temp */

	}
}

#if 0
void console::save_screen()
{
	FILE *out;
	unsigned int ptr, n;

	out = fopen("screen.bin","wb");
	if (!out) {
	  fprintf(stderr, "zce: error saving screen to `screen.bin`\n");
	  return;
	}

	ptr = (mem[0x400d] << 8) | mem[0x400c];
	n = 24*33 + 1;

	fwrite(mem+ptr,1,n,out);
	fclose(out);
}

void console::load_screen()
{
	FILE *in;
	unsigned int ptr, n;

	in = fopen("screen.bin","rb");
	if (!in) {
	  fprintf(stderr, "zce: error loading screen from `screen.bin`\n");
	  return;
	}

	ptr = (mem[0x400d] << 8) | mem[0x400c];
	n = 24*33 + 1;

	fread(mem+ptr,1,n,in);
	fclose(in);
}
#endif

void console::save_screen()
{
	FILE *out;
	unsigned int ptr, n;
        int c, cc;

	out = fopen("screen.txt","w");
	if (!out) {
	  fprintf(stderr, "zce: error saving screen to `screen.txt`\n");
	  return;
	}

	ptr = ((mem[0x400d] << 8) | mem[0x400c]) + 1;
        n = 0;

        while (n<24) {
          if (mem[ptr]==0x76) {
            fputc('\n',out);
            n++;
          } else {
            c = mem[ptr];
	    cc = zx2ascii[c&0x3f];
	    if (cc>='A' && cc<='Z' && !(c&0x80)) cc = tolower(cc);
	    fputc(cc,out);
          }
          ptr++;
        }

	fclose(out);
}

void console::load_screen()
{
	FILE *in;
	unsigned int ptr, n;
	int c;

	in = fopen("screen.txt","r");
	if (!in) {
	  fprintf(stderr, "zce: error loading screen from `screen.txt`\n");
	  return;
	}

	ptr = ((mem[0x400d] << 8) | mem[0x400c]) + 1;
	n = 0;

	while (n<24) {
	  c = fgetc(in);
          if (c==EOF) break;
	  if (c==10 && mem[ptr-1]!=0x76) {
            while (mem[ptr]!=0x76) ptr++;
	  } else if (c>=32) {
	    mem[ptr] = ascii2zxcs[c];
	    ptr++;
	  }
	  if (mem[ptr]==0x76) { ptr++; n++; }
	}

	fclose(in);
}
