#include <stdio.h>

/* use for example

./sixeltops < zx81.six |convert -rotate 90 -flatten -trim +repage - zx81.png

*/

#define NCOL 16

struct colorspec {
	int r, g, b;
};

static struct colorspec palette[NCOL];

void set_color(int c, int r, int g, int b)
{
	palette[c].r = r;
	palette[c].g = g;
	palette[c].b = b;
}

void initialize_palette()
{
	set_color(0, 0, 0, 0);
	set_color(1, 20, 20, 80);
	set_color(2, 80, 13, 13);
	set_color(3, 20, 80, 20);
	set_color(4, 80, 20, 80);
	set_color(5, 20, 80, 80);
	set_color(6, 80, 80, 20);
	set_color(7, 53, 53, 53);
	set_color(8, 26, 26, 26);
	set_color(9, 33, 33, 60);
	set_color(10, 60, 26, 26);
	set_color(11, 33, 60, 33);
	set_color(12, 60, 33, 60);
	set_color(13, 33, 60, 60);
	set_color(14, 60, 60, 33);
	set_color(15, 80, 80, 80);

}

void output_color(int c)
{
	printf("%4.2f %4.2f %4.2f setrgbcolor\n",
	       palette[c].r/100.,
	       palette[c].g/100.,
	       palette[c].b/100.);
}

int sixel()
{
	int c;

	do {
	  c = getchar();
	  if (c == EOF) return c;
	} while (c==0x0a || c==0x0d || c==0x20);

	return c;
}

int number(int *n)
{
	int c, k=0;

	for (;;) {
	  c = sixel();
	  if (c == EOF) return c;
	  if (c<0x30 || c>0x39) break;
	  k = k*10 + (c-0x30);
	}

	*n = k;

	return c;
}

int header()
{
	int c, p1=0, p2=0, p3=0;

	c = number(&p1);
	if (c == 0x3b) c = number(&p2);
	if (c == 0x3b) c = number(&p3);

	printf("%% header %d %d %d\n", p1, p2, p3);

	if (c != EOF) c = sixel();

	return c;
}

int escape()
{
	int c;

	c = sixel();
	if (c == EOF) return c;

	if (c == 0x50) {
	  c = header();
	  if (c == EOF) return c;
	} else if (c == 0x5c) {
	  return 0;
	}

	return c;
}

int raster()
{
	int c, pan=0, pad=0, ph=0, pv=0;

	c = number(&pan);
	if (c == 0x3b) c = number(&pad);
	if (c == 0x3b) c = number(&ph);
	if (c == 0x3b) c = number(&pv);

	printf("%% raster %d %d %d %d\n", pan, pad, ph, pv);

	return c;
}

int color()
{
	int c, r, g, b, n, d;

	c = number(&n);
	if (c == EOF) return c;

	if (c != 0x3b) {
	  output_color(n);
	  return c;
	}

	c = number(&d);
	if (c == 0x3b) c = number(&r); else return EOF;
	if (c == 0x3b) c = number(&g); else return EOF;
	if (c == 0x3b) c = number(&b); else return EOF;
	if (c == EOF) return c;

	set_color(n, r, g, b);
	output_color(n);

	return c;
}

int parse()
{
	int i, c, nr=1, xp=0, yp=0;

	c = sixel();

	for (;;) {

	  if (c==0 || c==EOF) return c;

	  if (c == 0x1b) {
	    c = escape();
	  } else if (c == 0x90) {
	    c = header();
	  } else if (c == 0x9c) {
	    return 0;
	  } else if (c == 0x21) {
	    c = number(&nr);
	  } else if (c == 0x22) {
	    c = raster();
	  } else if (c == 0x23) {
	    c = color();
	  } else if (c == 0x24) {
	    xp = 0;
	    c = sixel();
	  } else if (c == 0x2d) {
	    xp = 0;
	    yp += 6;
	    c = sixel();
	  } else {
	    c -= 63;
	    for (i=0; i<6; i++) if (c & (1<<i)) {
	      printf("newpath\n");
	      printf("%d %d moveto\n", yp+i, xp);
	      printf("%d %d lineto\n", yp+i, xp+nr);
	      printf("stroke\n");
	    }
	    xp += nr;
	    nr = 1;
	    c = sixel();
	  }

	}

}

int main()
{
	int c;

	initialize_palette();

	printf("%%!PS-Adobe-3.0\n");
	printf("%% sixeltops\n");

	c = parse();
	if (c == 0)
	  printf("showpage\n");
	else
	  printf("%% an error occurred\n");

	return 0;
}
