#include <stdio.h>
#include "z80.h"

extern unsigned int pcmask;
extern unsigned int in(int h,int l);
extern unsigned int out(int h,int l,int a);
extern void screen();
extern void inkey_ce();
extern void input_ce();
extern void save_p(int a);
extern void load_p(int a);
extern void lprint_ce(int ptr);
extern void remark_ce(int ptr);
extern void pause_ce(int n);
extern void scroll_ce();
extern void frames_ce();
extern void keyboard_ce();

unsigned long tstates=0;

#define parity(a) (partable[a])

unsigned char partable[256]={
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4
   };

void z80_processor::show(const int showpc)
{
	int i;
	unsigned char mask;

	if (showpc) printf("*PC: %04X  [PC]:   %02X\n", pc, mem[pc]);
	printf(" SP: %04X  [SP]: %02X%02X\n", sp, mem[sp+1], mem[sp]);
	printf("  A:   %02X     F:   %02X", a,f&0xd7);
	mask = 0x80;
	for (i=7; i>=0; i--) {
	  if (f & mask) switch (i) {
	    case 7 : printf(" S"); break;
	    case 6 : printf(" Z"); break;
	    case 4 : printf(" H"); break;
	    case 2 : printf(" P"); break;
	    case 1 : printf(" N"); break;
	    case 0 : printf(" C"); break;
	    default :;
	  }
	  mask >>= 1;
	}
	printf("\n");
	printf(" BC: %02X%02X  DE: %02X%02X    HL: %02X%02X\n", b,c, d,e, h,l);
/*
	printf("  A:   %02X     F:   %02X\n", a1,f1);
	printf(" BC: %02X%02X    DE: %02X%02X    HL: %02X%02X\n", b1,c1, d1,e1, h1,l1);
*/
	printf(" IX: %04X    IY: %04X\n", ix, iy);
}

void z80_processor::step(const int sverbose)
{
  static unsigned int daddr=0;

  do {

  ixoriy=new_ixoriy;
  new_ixoriy=0;
  intsample=1;
  op=fetch(pc&pcmask);

  if (pc!=daddr && pc&0x8000 && pcmask==0x17fff) {
    printf("zce: display information at %x %x %x\n",pc,daddr,pcmask);
    daddr = pc;
  }

  if (ixoriy==0) {

//  if ((verbose && sverbose) || pc>=0x10000)
  if (verbose && sverbose)
    printf("PC:%04X OP:%02X A:%02X F:%02X BC:%02X%02X DE:%02X%02X HL:%02X%02X SP:%04X FROM:%02X%02X\n", pc, op, a,f, b,c, d,e, h,l, sp, mem[sp+1], mem[sp]);

  if (op == 0x76) {
//    printf("halted\n");
    if (pc==bptr) cancel();
    return;
  }

  }

  pc++;
  radjust++;

  switch(op) {
#include "z80ops.c"
  }

  } while (new_ixoriy!=0);
}

void z80_processor::run(const unsigned int newpc)
{
//	printf("run %x\n",newpc);
	pc = newpc;

	while (!ctrlcd) {
	  step();
	  if (op == 0x76) return;
	}
}

void z80_processor::prepare()
{
	unsigned short tsp = sp-2;

	mem[0x4002] = (tsp & 0x00ff);
	mem[0x4003] = (tsp & 0xff00) >> 8;
}

void z80_processor::run_ret(const unsigned int newpc)
{
//	printf("run_ret %x\n",newpc);

	sp -= 2;
	store2b(sp,0x00,0x79); /* at halt */

	run(newpc);
}

void z80_processor::breakpoint(const unsigned int ptr)
{
        bptr = ptr;
	bins = mem[bptr];
	mem[bptr] = 0x76;
}

void z80_processor::cancel()
{
        if (bins==0x76) return;
	mem[bptr]=bins;
	bins=0x76;
}
