/* ZX81 emulator - command-line edition - by E. Olofsen */

#ifdef HAVE_GETLINE
#define _WITH_GETLINE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#ifdef HAVE_SELECT
#include <sys/select.h>
#include <termios.h>
#endif
#ifdef HAVE_SIGNAL
#include <signal.h>
#endif
#ifdef HAVE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#ifdef HAVE_SOCKET
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#endif
#ifdef HAVE_MXML
#include <zxml.h>
#endif

#include <string>
#include <sstream>
#include <iomanip>
#include "z80.h"
#include "zacvt.h"
#include "sixel.h"
#include "zline.h"
#include "console.h"

#ifdef DISASSEMBLER
#include "types.h"
#include "dissz80.h"
#endif

#ifdef HAVE_BMP
/*
 * Licence: Free to use, URL to my source and my name is required in your source code.
 *
 * Original source can be found at http://www.kalytta.com/bitmap.h
 *
 * Adapted for writing Bitcount=1 files
*/
#include "bitmap.h"
#endif

using namespace std;

#ifndef HAVE_GETLINE
#define STRLEN 256
#endif

/* global variables */

z80_processor z80;
static console cns;
FILE *sf=0, *ef=0;
unsigned char mem[MSIZE], meme[MSIZE], hrold[0x1800];
/* this two work on a per-k basis, so we can support 1k etc. properly */
unsigned char *memptr[MPSIZE], *memptre[MPSIZE];
int memattr[MPSIZE];
static int dotrace=0, dooutput=0;
static int docompile=0, dorun=0, dolist=0, doquit=0;
#ifdef HAVE_SIGNAL
static int doalarm=0;
#endif
int verbose=0, reroute=2, ctrlcd=0, instrtok=0;
extern unsigned long tstates;
unsigned int hrptr=0x2800;
static int uniquelines=0;
static ostringstream ossbuf;
static int possbuf=1;
static const int nscr=793;
static unsigned char memscr[nscr];
unsigned char hiressrc[0x1800];
unsigned int pcmask=0x1ffff;
static int fno=0;
unsigned char keyports[9]={0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff};
int xml_wrap_margin=0, xml_remtok=1, xml_binary=0;
static int sfloats=1;

#ifdef HAVE_BMP
static CBitmap cbmp;

void flipbmp()
{
  int i, j, k1, k2;
  unsigned char c;

  k1 = hrptr;
  k2 = k1+6112;
  for (j=0; j<96; j++) {
    for (i=0; i<32; i++) {
      c = mem[k1];
      mem[k1] = mem[k2];
      mem[k2] = c;
      k1++;
      k2++;
    }
    k2 -= 64;
  }
}

void writebmp(const char *filnam)
{
	unsigned int ptr1;
	RGBA *ptr2;
	int i, j, k, m;

	flipbmp();

	cbmp.CreateBits(256, 192);

	ptr1 = hrptr;
	ptr2 = (RGBA *) cbmp.GetBits();
	k = 0;
	for (i=0; i<0x1800; i++) {
	  m = 0x80;
	  for (j=0; j<8; j++) {
	    ptr2[k].Alpha = 0;
	    ptr2[k].Red = ptr2[k].Green = ptr2[k].Blue = (mem[ptr1] & m) ? 0 : 255;
	    k++;
	    m >>= 1;
	  }
	  ptr1++;
	}	  

	cbmp.Save(filnam,1);
}

void readbmp(const char *filnam)
{
	unsigned int ptr1;
	RGBA *ptr2;
	int i, j, k, m;

	if (!cbmp.Load(filnam)) return;

	if (cbmp.GetWidth()!=256 || cbmp.GetHeight()!=192) return;

	ptr1 = hrptr;
	ptr2 = (RGBA *) cbmp.GetBits();
	k = 0;
	for (i=0; i<0x1800; i++) {
	  m = 0x80;
	  mem[ptr1] = 255;
	  for (j=0; j<8; j++) {
	    if (ptr2[k].Red>=128 && ptr2[k].Green>=128 && ptr2[k].Blue>=128) mem[ptr1] &= ~m;
	    k++;
	    m >>= 1;
	  }
	  ptr1++;
	}	  

	flipbmp();

	hires();
}
#endif

#ifdef DISASSEMBLER
unsigned int disassemble(const unsigned int dAddr)
{
	DISZ80	*d;			/* Pointer to the Disassembly structure */
	int	err;

/* Allocate the dZ80 structure */
	d = (DISZ80*) malloc(sizeof(DISZ80));
	if (d == NULL)
		{
		printf("dz80: cannot allocate %ld bytes\n", sizeof(DISZ80));
		exit(1);
		}

	memset(d, 0, sizeof(DISZ80));

/* Set up dZ80's structure - it's not too fussy */
	memset(d, 0, sizeof(DISZ80));

/* Set the default radix and strings (comments and "db") */ 
	dZ80_SetDefaultOptions(d);		

/* Set the CPU type */
	d->cpuType = DCPU_Z80;

/* Set the start of the Z80's memory space */
	d->mem0Start = mem;

/* Indicate we're disassembling a single instruction */
	d->flags |= DISFLAG_SINGLE;

/* Set the disassembly address */
        d->start = d->end = dAddr;
		
	err = dZ80_Disassemble(d);
	if (err != DERR_NONE)
			{
			printf("**** dZ80 error:  %s\n", dZ80_GetErrorText(err));
			}
		
/* Display the disassembled line, using the hex dump and disassembly buffers in the DISZ80 structure */
	printf(" PC: %04X: %10s:  %s\n", dAddr, d->hexDisBuf, d->disBuf);
		
	free(d);

	return dAddr+(unsigned int)d->bytesProcessed;
}
#endif

#ifndef HAVE_MEMRCHR

char *memrchr(char *s, char c, size_t n)
{
	int i;
	for (i=n-1; i>=0; i--) if (s[i]==c) return s+i;
	return NULL;
}

#endif

#ifdef HAVE_SELECT

static struct termios st = {0};

void initts()
{
	if (doquit) return;

        if (tcgetattr(0, &st) < 0) perror("tcgetattr()");
        st.c_lflag &= ~ICANON;
        st.c_lflag &= ~ECHO;
        st.c_cc[VMIN] = 1;
        st.c_cc[VTIME] = 0;
        if (tcsetattr(0, TCSANOW, &st) < 0) perror("tcsetattr ICANON");
}

void resetts()
{
	if (doquit) return;

        st.c_lflag |= ICANON;
        st.c_lflag |= ECHO;
        if (tcsetattr(0, TCSANOW, &st) < 0) perror ("tcsetattr ~ICANON");
}

char getch() {
        char c=0;
	int n;
	struct timeval tv = { 0L, 0L };
	fd_set fds;

	FD_ZERO(&fds);
	FD_SET(0, &fds);

        if (select(1, &fds, NULL, NULL, &tv)>0) {
          n = read(0, &c, 1);
	  if (n<0) perror("read()");
	}

  int b, y;
  char cc;

  cc = tolower(c);

for(y=0;y<8;y++)		/* 8 half-rows */
  {
  b=0;	/* we set bits in b as appropriate */
  switch(y)	/* below comments given in order b1->b5 */
    {
    /* left-hand side */
    case 0:	/* sft,z,x,c,v */
      if ((c>='A' && c<='Z') || c=='!' || c=='@' || c=='#' || c=='$' || c=='%' || c=='^' || c=='&' || c=='*' || c=='(' || c==')' || c=='>' || c=='\'' || c=='`') b|=1;
      if(cc=='z') b|=2;
      if(cc=='x') b|=4;
      if(cc=='c') b|=8;
      if(cc=='v') b|=16;
      break;
    case 1:	/* a,s,d,f,g */
      if(cc=='a') b|=1;
      if(cc=='s') b|=2;
      if(cc=='d') b|=4;
      if(cc=='f') b|=8;
      if(cc=='g') b|=16;
      break;
    case 2:	/* q,w,e,r,t */
      if(cc=='q') b|=1;
      if(cc=='w') b|=2;
      if(cc=='e') b|=4;
      if(cc=='r') b|=8;
      if(cc=='t') b|=16;
      break;
    case 3:	/* 1,2,3,4,5 */
      if(c=='1' || c=='!') b|=1;
      if(c=='2' || c=='@') b|=2;
      if(c=='3' || c=='#') b|=4;
      if(c=='4' || c=='$') b|=8;
      if(c=='5' || c=='%') b|=16;
      break;

    /* right-hand side */
    case 4:	/* 0,9,8,7,6 */
      if(c=='0' || c==')') b|=1;
      if(c=='9' || c=='(') b|=2;
      if(c=='8' || c=='*') b|=4;
      if(c=='7' || c=='&') b|=8;
      if(c=='6' || c=='^') b|=16;
      break;
    case 5:	/* p,o,i,u,y */
      if(cc=='p') b|=1;
      if(cc=='o') b|=2;
      if(cc=='i') b|=4;
      if(cc=='u') b|=8;
      if(cc=='y') b|=16;
      break;
    case 6:	/* ent,l,k,j,h */
      if(c==10 || c=='\'') b|=1;
      if(cc=='l') b|=2;
      if(cc=='k') b|=4;
      if(cc=='j') b|=8;
      if(cc=='h') b|=16;
      break;
    case 7:	/* spc,dot,m,n,b */
      if(c==' ' || c=='`') b|=1;
      if(c=='.' || c=='>') b|=2;
      if(cc=='m') b|=4;
      if(cc=='n') b|=8;
      if(cc=='b') b|=16;
      break;
    }
  keyports[y]=((b^31)|0xe0);	/* some things need top bits to be 1 */
  }
        return c;
}

#endif

unsigned int getstring(unsigned char vr)
{
	vr = ascii2zx[vr];

	mem[0x10017] = vr;
	mem[0x10018] = ascii2zx['$'];
	mem[0x10019] = 0x76;
	mem[0x4001] |= 0x80; /* run */

	mem[0x4016] = 0x17; /* ch-add */
	mem[0x4017] = 0x00;

        meme[0x001b] = meme[0x0050] = 1;
	z80.run_ret(0x111c); /* look-vars */
        meme[0x001b] = meme[0x0050] = 0;

	if (z80.f&0x01) {
	  return 0x10017;
	} else {
	  return ((z80.h<<8) | z80.l);
	}
}

void setstring(unsigned char vr, const char *s)
{
	unsigned int ptr;
	int i, n, ns, nc;

	if (s) ns=strlen(s); else ns=0;

	vr = ascii2zx[vr];

	mem[0x10017] = vr;
	mem[0x10018] = ascii2zx['$'];
	mem[0x10019] = 0x76;

	mem[0x4001] |= 0x80; /* run */

	mem[0x4016] = 0x17; /* ch-add */
	mem[0x4017] = 0x00;

        meme[0x001b] = meme[0x0050] = 1;
	z80.run_ret(0x111c); /* look-vars */
        meme[0x001b] = meme[0x0050] = 0;

	if (z80.f&0x01) {
	  z80.h = mem[0x4011];
	  z80.l = mem[0x4010];
	  nc = ns + 3;
	  z80.b = (nc & 0xff00) >> 8;
	  z80.c = nc & 0x00ff;
	  z80.run_ret(0x99e);
	  ptr = (mem[0x4011]<<8) | mem[0x4010];
	} else {
	  ptr = (z80.h<<8) | z80.l;
	  n = (mem[ptr+2]<<8) | mem[ptr+1];
	  n = ns - n;
	  z80.b = (n & 0xff00) >> 8;
	  z80.c = n & 0x00ff;
	  if (n>0) {
	    z80.run_ret(0x099e);
	    z80.b = (n & 0xff00) >> 8;
	    z80.c = n & 0x00ff;
	  } else if (n<0) {
	    n = -n;
	    z80.b = (n & 0xff00) >> 8;
	    z80.c = n & 0x00ff;
	    z80.run_ret(0x0a60);
	  }
	}
	mem[ptr++] = (vr & 0x5f) | 0x40;
	mem[ptr++] = ns & 0x00ff;
	mem[ptr++] = (ns & 0xff00) >> 8;
	for (i=0; i<ns; i++) mem[ptr++] = ascii2zxcs[(int)s[i]];
}

#ifdef HAVE_SOCKET

int sockmsg(unsigned int ptr, int sl=-1)
{
    int sockfd, portno, n;
    struct sockaddr_in serv_addr;
    struct hostent *server;
    char buffer[256];
    unsigned char c, cc;

    portno = 8181;
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) {
        perror("zce: error opening socket");
	return 1;
    }
    server = gethostbyname("localhost");
    if (server == NULL) {
        fprintf(stderr,"zce: no such host\n");
        return 1;
    }
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    bcopy((char *)server->h_addr, 
         (char *)&serv_addr.sin_addr.s_addr,
         server->h_length);
    serv_addr.sin_port = htons(portno);
    if (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) {
        perror("zce: error connecting");
        return 1;
    }

    n = 0;
    while (1) {
      if (n==sl) break;
      c = mem[ptr++];
      if (c==0x76) break;
      cc = zx2ascii[c&0x3f];
      if (reroute==0 && cc>='A' && cc<='Z' && !(c&0x80))
	cc = tolower(cc);
      buffer[n++] = cc;
    }
    buffer[n] = '\0';

    n = write(sockfd,buffer,strlen(buffer));
    if (n < 0) {
         perror("zce: error writing to socket");
	 return 1;
    }
    bzero(buffer,256);
    n = read(sockfd,buffer,255);
    if (n < 0) {
         perror("zce: error reading from socket");
         return 1;
    }
    setstring('I',buffer);
    close(sockfd);
    return 0;
}

int httpc(unsigned int ptr, int sl)
{
    const int nbuf=2048, nbufchnk=32;
    int sockfd, portno, n, offset;
    struct sockaddr_in serv_addr;
    struct hostent *server;
    char *buffer, *request;
    unsigned char c, cc;
    int fd, cnt=0, bpos=0;

    portno = 80;
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) {
        perror("zce: error opening socket");
	return 1;
    }

    buffer = (char *) malloc(nbuf);

    n = 0;
    while (1) {
      if (n==sl) break;
      c = mem[ptr++];
      if (c==0x76) break;
      cc = zx2ascii[c&0x3f];
      cc = tolower(cc);
      if (cc!=':') {
        buffer[n++] = cc;
      } else {
        buffer[n++] = '\0';
        offset = n;
      }
    }
    buffer[n] = '\0';

    server = gethostbyname(buffer);
    if (server == NULL) {
        fprintf(stderr,"zce: no such host\n");
	free(buffer);
        return 1;
    }
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    bcopy((char *)server->h_addr, 
         (char *)&serv_addr.sin_addr.s_addr,
         server->h_length);
    serv_addr.sin_port = htons(portno);
    if (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) {
        perror("zce: error connecting");
	free(buffer);
        return 1;
    }

    request = (char *) malloc(256);
    sprintf(request, "GET %s HTTP/1.1\r\nHost: zx81\r\nConnection: close\r\n\r\n", buffer+offset);
    n = write(sockfd,request,strlen(request));
    free(request);

    if (n < 0) {
        perror("zce: error writing to socket");
	free(buffer);
	return 1;
    }

    fd = open("download",O_WRONLY|O_CREAT|O_TRUNC,0600);

    for (cnt=0; ;cnt++) {

    n = read(sockfd,buffer+bpos,nbufchnk);
    if (n < 0) {
        perror("zce: error reading from socket");
	free(buffer);
	close(fd);
        return 1;
    }

    if (n==0) {
      buffer[bpos+n] = '\0';
      setstring('I',buffer);
      break;
    } else {
      write(fd,buffer+bpos,n);
    }

    bpos += n;

    if (bpos>=nbuf-nbufchnk) break;

    }

    free(buffer);
    close(fd);
    close(sockfd);
    return 0;
}

#endif

unsigned int in(int h,int l)
{
   if (verbose) {
   printf(" [%ld] in at %4x: %2x %2x\n", tstates, z80.pc, h, l);

   switch(h)
      {
      case 0xfe:	printf("%x\n",keyports[0]); break;
      case 0xfd:	printf("%x\n",keyports[1]); break;
      case 0xfb:	printf("%x\n",keyports[2]); break;
      case 0xf7:	printf("%x\n",keyports[3]); break;
      case 0xef:	printf("%x\n",keyports[4]); break;
      case 0xdf:	printf("%x\n",keyports[5]); break;
      case 0xbf:	printf("%x\n",keyports[6]); break;
      case 0x7f:	printf("%x\n",keyports[7]); break;
      }
   }

   switch(h)
      {
      case 0xfe:	return(keyports[0]);
      case 0xfd:	return(keyports[1]);
      case 0xfb:	return(keyports[2]);
      case 0xf7:	return(keyports[3]);
      case 0xef:	return(keyports[4]);
      case 0xdf:	return(keyports[5]);
      case 0xbf:	return(keyports[6]);
      case 0x7f:	return(keyports[7]);
      }

   return 0;
}

unsigned int out(int h,int l,int a)
{
	if (verbose) printf(" [%ld] out at %4x: %2x %2x %2x\n", tstates, z80.pc, h, l, a);

	if (l==0xf0) printf("%c", a);

	return 0;
}

int atoih(const char *s)
{
	int n;
	sscanf(s,"%i",&n);
	return n;
}

void put_keyboard()
{
#ifdef HAVE_SELECT
	unsigned int kv=0;
	char c;

	if (mem[0x403b] & 0x01) {
	  kv = 0xffff;
	  mem[0x403b] &= ~0x01;
	} else {
          c = getch();
          switch (c) {
            case '1' : kv = 0xfdf7; break;
            case '2' : kv = 0xfbf7; break;
            case '3' : kv = 0xf7f7; break;
            case '4' : kv = 0xeff7; break;
            case '5' : kv = 0xdff7; break;

            case '6' : kv = 0xdfef; break;
            case '7' : kv = 0xefef; break;
            case '8' : kv = 0xf7ef; break;
            case '9' : kv = 0xfbef; break;
            case '0' : kv = 0xfdef; break;

            case 'q' : kv = 0xfdfb; break;
            case 'w' : kv = 0xfbfb; break;
            case 'e' : kv = 0xf7fb; break;
            case 'r' : kv = 0xeffb; break;
            case 't' : kv = 0xdffb; break;

            case 'y' : kv = 0xdfdf; break;
            case 'u' : kv = 0xefdf; break;
            case 'i' : kv = 0xf7df; break;
            case 'o' : kv = 0xfbdf; break;
            case 'p' : kv = 0xfddf; break;

            case 'a' : kv = 0xfdfd; break;
            case 's' : kv = 0xfbfd; break;
            case 'd' : kv = 0xf7fd; break;
            case 'f' : kv = 0xeffd; break;
            case 'g' : kv = 0xdffd; break;

            case 'h' : kv = 0xdfbf; break;
            case 'j' : kv = 0xefbf; break;
            case 'k' : kv = 0xf7bf; break;
            case 'l' : kv = 0xfbbf; break;
	    case 0x0d : kv = 0xfdbf; break;

            case 'z' : kv = 0xfbfe; break;
            case 'x' : kv = 0xf7fe; break;
            case 'c' : kv = 0xeffe; break;
            case 'v' : kv = 0xdffe; break;

            case 'b' : kv = 0xdf7f; break;
            case 'n' : kv = 0xef7f; break;
            case 'm' : kv = 0xf77f; break;
            case '.' : kv = 0xfb7f; break;
            case ' ' : kv = 0xfd7f; break;
/*
            case 'A' : kv = 0xfcfd; break;
            case 'B' : kv = 0xde7f; break;
*/
            default  :;
          }
	  mem[0x403b] |= 0x01;
	}
	if (kv) {
          mem[0x4025] = kv & 0x00ff;
          mem[0x4026] = (kv & 0xff00) >> 8;
	}
#endif
}

void screen()
{
	unsigned int ptr;
	int n = 0;
	char c, cc;
	ptr = mem[0x400c] + (mem[0x400d] << 8);
	for (;;) {
	  c = mem[ptr];
	  if (c==0x76) {
	    putchar('\n');
	    n++;
	    if (n>24) break;
	   } else {
	    cc = zx2ascii[c&0x3f];
	    if (reroute==0 && !(c&0x80)) cc = tolower(cc);
	    putchar(cc);
	   }
	  ptr++;
	}
}

void hcopy()
{
	FILE *f;
	unsigned int ptr;
	int n = 1;
	unsigned char c, cc;

//	f = fopen("screen.html","w");
	f = stdout;

	fprintf(f,"<!DOCTYPE HTML>\n");
	fprintf(f,"<html xmlns=\"http://www.w3.org/1999/xhtml\">\n");
	fprintf(f,"  <head>\n");
	fprintf(f,"    <title>ZX81</title>\n");
	fprintf(f,"    <link href=\"zx.css\" rel=\"stylesheet\" type=\"text/css\"/>\n");
	fprintf(f,"  </head>\n");
	fprintf(f,"<body>\n");
	fprintf(f,"<pre>\n");

	ptr = mem[0x400c] + (mem[0x400d] << 8) + 1;
	for (;;) {
	  c = mem[ptr];
	  if (c==0x76) {
	    fprintf(f,"\n");
	    n++;
	    if (n>24) break;
	   } else {
	    if (c>=0x01 && c<=0x0a) {
	      fprintf(f,"&#%d;",c+176);
	    } else if (c>=0x80 && c<= 0x90) {
	      fprintf(f,"&#%d;",c+64);
	    } else {
	      cc = zx2ascii[c&0x3f];
	      if (cc>='A' && cc<='Z' && !(c&0x80)) cc = tolower(cc);
	      fprintf(f,"%c",cc);
	    }
	   }
	  ptr++;
	}

	fprintf(f,"</pre>\n");
	fprintf(f,"</body>\n");
	fprintf(f,"</html>\n");

//	fclose(f);
}

void load(const char *filnam)
{
	FILE *in;

	if (verbose) fprintf(stderr, "loading `%s`\n", filnam);

	in = fopen(filnam,"rb");
	if (!in) {
	  fprintf(stderr, "zce: error loading `%s`\n", filnam);
	  return;
	}
	fread(mem+0x4009,1,0x10000-0x4009,in); // range check
	fclose(in);

return;

if (!sfloats) {
  unsigned int ptr, ptrl, ptre, llen, k;
  ptre = (mem[0x400d] << 8) | mem[0x400c];
  ptr = 0x407d;

  while (ptr<ptre) {
    k = 0;
    ptr += 2;
    llen = (mem[ptr+1] << 8) | mem[ptr];
    ptr += 2;
    if (mem[ptr]==0xea) {
      ptr += llen;
    } else {
      ptrl = ptr+llen;
      ptr++;
      for (; ptr<ptrl; ptr++) {
        if (mem[ptr]==0x7e) k=6;
        if (k>0) { mem[ptr]=0x7f; k--; }
      }
    }
  }
}

}

void save(const char *filnam)
{
	FILE *out;
	unsigned int n;

	if (verbose) fprintf(stderr, "saving `%s`\n", filnam);

	out = fopen(filnam,"wb");
	if (!out) {
	  fprintf(stderr, "zce: error saving `%s`\n", filnam);
	  return;
	}

	n = (mem[0x4015] << 8) | mem[0x4014];
	n -= 0x4009;

        mem[0x4035] |= 0x80; // avoid crash in real emulator/system

        mem[0x4028] = 55;
        mem[0x4029] = 0x7d;
	mem[0x4030] = 0x40;
        mem[0x4035] |= 0x80;
        mem[0x403a] = 24;
        mem[0x403b] = 0x40;

	fwrite(mem+0x4009,1,n,out);
	fclose(out);
}

void save_p(int a)
{
	string s;
	char c;

	if (a <= 0x8000) {  // range check
	  do {
	    c = mem[a++];
	    s += (char) tolower(zx2ascii[c&0x3f]);
	  } while (!(c&0x80));
	  s += ".p";
	} else {
	  s = "zx81.p";
	}

	save(s.c_str());
}

void load_p(int a)
{
	string s;
	char c;

	if (a <= 0x8000) { // range check
	  do {
	    c = mem[a++];
	    s += (char) tolower(zx2ascii[c&0x3f]);
	  } while (!(c&0x80));
	  s += ".p";
	} else {
	  s = "zx81.p";
	}

	load(s.c_str());
}

void load_binary(const char *filnam)
{
	FILE *in;

	if (verbose) fprintf(stderr, "loading `%s`\n", filnam);

	in = fopen(filnam,"rb");
	if (!in) {
	  fprintf(stderr, "zce: error loading `%s`\n", filnam);
	  return;
	}
//	fread(mem+0x10000,1,0x10000,in);
	fread(mem+0x2000,1,0x1800,in);
	fclose(in);
}

void getgdb(FILE *in, const int blen, const int fn, const int fno)
{
  unsigned char *m;
  int i;
  unsigned int ptr;
  m = (unsigned char *) malloc(blen);
  fread(m,1,blen,in);
  if (!fno || fn==fno) printf("%2d: ",fn);
  for (i=88;;i++) {
    if (!fno || fn==fno) printf("%c",zx2ascii[m[i]&0x3f]);
    if (m[i]&0x80) break;
  }
  if (!fno || fn==fno) printf("\n");
  if (fn==fno) {
    ptr = 0x4009;
    for (i++; i<blen; i++) mem[ptr++] = m[i];
  }
  free(m);
}

void loadtzx(const char *filnam, const int fno)
{
	FILE *in;
        int c;
        int parsing=1, state=0, cnt=0, blen, fn=0;

	in = fopen(filnam,"rb");
	if (!in) {
	  fprintf(stderr, "zce: error loading `%s`\n", filnam);
	  return;
	}

        while (parsing) {
          c = fgetc(in);
          if (c<0) break;
          switch (state) {
            case 0   : if (c==26) { printf("\n"); state++; break; }
                       putchar(c);
		       break;
	    case 1   : printf("Version: %d",c);
		       c = fgetc(in);
		       printf(".%d\n",c);
		       state++;
		       break;
	    case 2   : if (verbose) printf("Block ID: %d\n",c);
		       state = c;
		       break;
	    case 0x19: blen = c;
		       c = fgetc(in);
		       blen = (c<<8) | blen;
		       if (verbose) printf("Length: %d\n", blen);
		       c = fgetc(in);
		       c = fgetc(in);
		       getgdb(in,blen,++fn,fno);
		       state = 2;
		       break;
            case 0x30: if (cnt==0) {
		         cnt = c;
			 if (verbose) printf("Length: %d; \"",cnt);
		       } else {
		         putchar(c);
		         cnt--;
		         if (cnt==0) {
		           printf("\"\n");
		           state = 2;
		         }
		       }
		       break;
            default  : parsing = 0;
          }
        }

        fclose(in);
}

void delete_top(const int n)
{
	if (!n) return;

	z80.l = n & 0x00ff;
	z80.h = (n & 0xff00) >> 8;
	z80.run_ret(0x09d8); /* line-addr */
	z80.d = z80.h;
	z80.e = z80.l;
	z80.h = mem[0x400d];
	z80.l = mem[0x400c];
	z80.run_ret(0x0a5d); /* reclaim */
}

void compile(const char *filnam)
{
	FILE *fp;
#ifdef HAVE_GETLINE
	char *line = NULL;
        size_t linecap = 0;
#else
	char  line[STRLEN];
#endif
        ssize_t linelen;
	zline zl;
	int add=0;

	fp = fopen(filnam,"r");
	if (!fp) {
	  fprintf(stderr, "zce: error loading `%s`\n", filnam);
	  return;
	}

	while (1) {
#ifdef HAVE_GETLINE
          if ((linelen = getline(&line, &linecap, fp)) <= 0) break;
#else
	  if (!fgets(line,STRLEN-1,fp)) break;
	  linelen = strlen(line);
#endif
	  if (linelen>1) if (line[0]==' ' || (line[0]>='0' && line[0]<='9') || add) {
	    line[linelen-1] = '\0';
	    if (line[linelen-2]=='\\') {
	      line[linelen-2] = '\0';
	      if (add) zl.add(line); else zl.set(line);
	      add = 1;
	    } else {
	      if (add) zl.add(line); else zl.set(line);
	      add = 0;
	      zl.compile();
	      if (zl.insert(uniquelines)) break;
	    }
	  }
	}

	fclose(fp);
}

int checkscr()
{
  unsigned int ptr = mem[0x400c] + (mem[0x400d] << 8);
  int chgd=0, empty=1;
  int i;

  for (i=0; i<nscr; i++) {
    if (mem[ptr+i]!=0 && mem[ptr+i]!=0x76) empty=0;
    if (memscr[i] != mem[ptr+i]) {
      chgd = 1;
      memscr[i] = mem[ptr+i];
    }
  }
  return chgd && !empty;
}

int checkhrscr()
{
  unsigned int ptr = hrptr;
  int chgd=0, empty=1;
  int i;

  for (i=0; i<0x1800; i++) {
    if (!mem[ptr+i]) empty=0;
    if (hrold[i] != mem[ptr+i]) {
      chgd = 1;
      hrold[i] = mem[ptr+i];
    }
  }
  return chgd && !empty;
}

void inkey_ce()
{
	usleep(10000);

	z80.b = z80.c = 0;

#ifdef HAVE_SELECT
	char c;
	c = getch();
	if (c) {
	  if (c==4) ctrlcd = 1;
	  z80.b = 0; z80.c = 1;
	  c = ascii2zx[(int)c];
	  mem[0x401e] = c; /* berg */
	  z80.d = 0x40;
	  z80.e = 0x1e;
	}
#endif
}

void input_ce()
{
#ifdef HAVE_GETLINE
	char *line = NULL;
        size_t linecap = 0;
#else
	char line[STRLEN];
#endif
        ssize_t linelen;
	unsigned int ptr, sptr;
	int i;
	int ns;
        string s;
        zline zl;

#ifdef HAVE_SELECT
	resetts();
#endif

	ns = mem[0x4001] & 0x40; /* numerical or string */

	if (!doquit) {

	if (reroute==0) {
	  printf("> ");
	} else {
	  if (ns)
	    printf("L> ");
	  else
	    printf("L$> ");
	}

	}

	do {

#ifdef HAVE_GETLINE
        linelen = getline(&line, &linecap, stdin);
	if (linelen<0) {
	  ctrlcd = 1;
	  return;
	}
	linelen--;
	line[linelen] = '\0';
#else
	if (!gets(line)) {
	  ctrlcd = 1;
	  return;
	}
	linelen = strlen(line);
#endif

	} while (line[0]=='#') ;

        if (ns) {
          s=line;
        } else {
          s='\"'; s+=line; s+='\"';
          linelen += 2;
        }

	z80.b = 0;
	z80.c = ns ? linelen+1 : linelen+3;

	z80.run_ret(0x0030); /* bc-spaces */
	ptr = (z80.d << 8) | z80.e;

/* copy the line to the workspace */
	sptr = ptr;
        if (instrtok) {
          zl.set(s.c_str());
          zl.compile(4);
	  for (i=0; i<(int)zl.length(); i++) mem[ptr++] = zl[i];
        } else {
	  if (reroute==0)
	    for (i=0; i<linelen; i++) mem[ptr++] = ascii2zxcs[(int)s[i]];
	  else
	    for (i=0; i<linelen; i++) mem[ptr++] = ascii2zx[(int)s[i]];
        }
	mem[ptr] = 0x76;

	z80.run_ret(0x14bc); /* set-mem */
	mem[0x4000] = 0xff; /* clear error */

	mem[0x4001] &= ~0x80; /* syntax check */
	mem[0x4016] = sptr & 0x00ff; /* ch-add */
	mem[0x4017] = (sptr & 0xff00) >> 8;
	z80.run_ret(0x0f55); /* scanning */

	if (mem[0x4000] == 0xff) {
	  mem[0x4001] |= 0x80; /* run */
	  mem[0x4016] = sptr & 0x00ff; /* ch-add */
	  mem[0x4017] = (sptr & 0xff00) >> 8;
	  z80.run_ret(0x0d6b); /* input */
	}

	z80.run_ret(0x14a3); /* x-temp */

#ifdef HAVE_SELECT
	initts();
#endif
}

void lprint_ce(int ptr)
{
	static int cnt=1, lfout=1, scnt=0;
	unsigned char c, cc;

	while (1) {
	  c = mem[ptr];
	  if (c == 0x76) break;
	  if (lfout && verbose) { ossbuf << setw(4) << cnt << ": "; lfout=0; }
	  if (reroute==0 && ((c>=0x01 && c<=0x0a) || (c>=0x80 && c<=0x8a))) {
	    ossbuf << "\"+chr$(" << (int)c << ")+\"";
	  } else {
	    if (c==0x80) cc = '~';
	    else cc = zx2ascii[c&0x3f];
	    if (reroute==0 && cc>='A' && cc<='Z' && !(c&0x80))
	      cc = tolower(cc);
	    if (cc==' ') {
	      scnt++;
	    } else {
	      if (scnt) {
	        for (int i=1; i<=scnt; i++) ossbuf << ' ';
	        scnt = 0;
	      }
	      if (c==0xb8 && !(mem[0x4001]&0x80)) ossbuf << '<';
	      ossbuf << cc;
	    }
	    if (c==0xb8 && !(mem[0x4001]&0x80)) ossbuf << '>';
	  }
	  ptr++;
	}
	if (z80.a==0x76) { /* trick to avoid wrapping */
	  if (possbuf) {
	    printf("%s\n", ossbuf.str().c_str());
	    ossbuf.str("");
	  }
	  cnt++;
	  lfout = 1;
	  scnt = 0;
	}
}

void dohex(int dptr)
{
	char c, c1, c2;
	int ptr=dptr, hptr=dptr;
	int ll;

        c = zx2ascii[mem[++hptr]&0x3f];
	if (c!='H') return;
        c = zx2ascii[mem[++hptr]&0x3f];
	if (c!='E') return;
        c = zx2ascii[mem[++hptr]&0x3f];
	if (c!='X') return;
        c = zx2ascii[mem[++hptr]&0x3f];
	if (c!=':') return;
		
	while (1) {
          c1 = mem[++hptr];
	  if (c1<28 || c1>43) break;
          c2 = mem[++hptr];
	  if (c2<28 || c2>43) break;
          hptr++;
	  if (mem[hptr]!=26) break;
	  mem[ptr] = ((c1-28) << 4) | (c2-28);
	  ptr++;
	}

	hptr++;
	z80.d = (ptr & 0xff00) >> 8;
	z80.e = ptr & 0x00ff;
	z80.h = (hptr & 0xff00) >> 8;
	z80.l = hptr & 0x00ff;
	z80.run_ret(0x0a5d); /* reclaim */

	ll = (mem[dptr-2]<<8) + (mem[dptr-3]) - (hptr-ptr);
	mem[dptr-2] = (ll & 0xff00) >> 8;
	mem[dptr-3] = ll & 0x00ff;

	ptr++;
	mem[0x4029] = (ptr & 0x00ff); /* nxtlin */
	mem[0x402a] = (ptr & 0xff00) >> 8;
}

void remark_ce(int dptr)
{
	char c, vr;
	int i, ptr;
	int ns, ns3;
	char *es;
	char ev[4] = {'Z','X',' ','\0'};
	string s;

	if (!(mem[0x4001] & 0x80)) return;

        c = zx2ascii[mem[dptr]&0x3f];
#ifdef DEBUG
	printf("remark %x %x %c\n", dptr, mem[dptr], c);
#endif
        if (c=='(') {
	  dohex(dptr);
	  return;
	} else if (c=='$') {
	  i = 1;
	  while (mem[dptr+i]!=0x76)
	    s = s + (char)tolower(zx2ascii[mem[dptr+i++]&0x3f]);
	  ns = s.length();
	  if (ns>2) {
	    switch (s[0]) {
	      case 'b' : s.erase(0,2);
			 load_binary(s.c_str());
			 break;
	      case 'd' : s.erase(0,2);
			 delete_top(atoi(s.c_str()));
			 break;
	      case 'm' : s.erase(0,2);
			 compile(s.c_str());
			 break;
	      case 'p' : dptr += 2;
			 mem[0x4016] = dptr & 0x00ff; /* ch-add */
			 mem[0x4017] = (dptr & 0xff00) >> 8;
			 mem[0x4001] |= 0x80; /* run */
			 z80.run_ret(0x0d92); /* class-6 */
			 z80.run_ret(0x0020); /* next char */
			 z80.run_ret(0x0d92); /* class-6 */
			 z80.run_ret(0x0bf5); /* stk-to-bc */
			 hires_plot(z80.c,z80.b);
			 z80.sp += 2;
			 break;
	      case 'r' : s.erase(0,2);
			 rxml(s.c_str());
			 break;
	      case 'w' : s.erase(0,2);
			 wxml(s.c_str());
			 break;
	      default  :;
	    }
	    return;
	  }
	  if (ns<1) return;
	  if (s[1]!='$') {
	    switch (s[0]) {
#ifdef HAVE_SOCKET
/*
	      case 'i' : ptr = getstring('I');
			 if (ptr==0xfff4)
			   sockmsg(ptr);
			 else
			   sockmsg(ptr+3,(mem[ptr+2]<<8)|mem[ptr+1]);
			 break;
*/
	      case 'i' : ptr = getstring('I');
			 httpc(ptr+3,(mem[ptr+2]<<8)|mem[ptr+1]);
			 break;
#endif
	      case 'h' : hires();
			 break;
	      case 'c' : hires_clear();
			 break;
	      case 'k' : instrtok = !instrtok;
			 break;
	      case 's' : if (checkscr()) switch (reroute) {
		           case 3 : hcopy(); break;
			   case 2 : sixel_copy(); break;
			   default: screen();
			 }
			 break;
	      case 't' : dotrace = !dotrace;
			 break;
	      case 'w' : hcopy();
			 break;
	      case '<' : cns.load_screen();
			 break;
	      case '>' : cns.save_screen();
			 break;
	      default  :;
	    }
	    return;
	  }

	  mem[dptr+1] &= 0x7f;
	  vr = mem[dptr+1];

	  z80.run_ret(0x0049); /* increment ch-add */
	  mem[0x4001] |= 0x80; /* run instead of syntax check */
	  z80.run_ret(0x111c); /* get variable information */

	  ev[2] = zx2ascii[vr&0x3f];

	  if (z80.f&0x01) {
/* assign to basic variable */
	    es = getenv(ev);
	    if (!es) return;
	    ns = strlen(es);
	    z80.h = mem[0x4011]; /* vars */
	    z80.l = mem[0x4010];
	    ns3 = ns+3;
	    z80.b = (ns3 & 0xff00) >> 8;
	    z80.c = ns3 & 0x00ff;
	    z80.run_ret(0x99e); /* make-room */
	    ptr = (mem[0x4011]<<8) | mem[0x4010];
	    mem[ptr++] = (vr & 0x5f) | 0x40;
	    mem[ptr++] = ns & 0x00ff;
	    mem[ptr++] = (ns & 0xff00) >> 8;
	    for (i=0; i<ns; i++) mem[ptr++] = ascii2zxcs[(int)es[i]];
	  } else {
/* export to environment variable */
	    ptr = (z80.h << 8) | z80.l;
	    ptr++;
	    ns = (mem[ptr+1] << 8) | mem[ptr];
	    ptr+=2;
	    s = "";
	    for (i=0; i<ns; i++) s += zx2ascii[mem[ptr++]&0x3f];
	    if (!ef) ef = fopen("zx81.env","w");
	    fprintf(ef,"export %s='%s'\n", ev, s.c_str());
	  }
	}
}

void pause_ce(int n)
{
	int n1, n2;

	n1 = n / 50;
	n2 = n - n1*50;

	if (n1>0) sleep(n1);
	if (n2>0) usleep(n2*20000);
}

/* for the sixel output routine the screen has to remain rectangular */

void scroll_ce()
{
	unsigned int ptr;
	int i, j;

	z80.b = mem[0x4022]+1;
	z80.c = 0x21;
	z80.run_ret(0x918);

	ptr = (mem[0x400d] << 8) | mem[0x400c];

	for (j=1; j<23; j++) {
	  for (i=1; i<=32; i++) {
	    mem[ptr+i] = mem[ptr+i+33];
	  }
	  ptr += 33;
	}

}

void frames_ce()
{
	int r = rand();
	mem[0x4034] = r & 0x00ff;
	mem[0x4035] = (r & 0xff00) >> 8;
}

void keyboard_ce()
{
	if (verbose) printf("routine keyboard called\n");

	usleep(10000);

        z80.h = z80.l = 0xff;
        z80.run_ret(0x02be);
        z80.pc = 0x02bd;
        z80.op = 0x00;

        for (int i=0; i<9; i++) keyports[i] = 0xff;
}

void theclock()
{
	time_t tv;
	struct tm *tms;
	time(&tv);
	tms = localtime(&tv);
	mem[0x10001] = tms->tm_hour;
	mem[0x10002] = tms->tm_min;
	mem[0x10003] = tms->tm_sec;
}

void initmem()
{
int f;
int count;

memset(mem+0x4000,0,MSIZE-0x4000);
memset(meme,0,MSIZE);

/* ROM setup */
count=0;
for(f=0;f<16;f++)
  {
  if (f<8)
    memattr[f]=0;
  else
    memattr[f]=1;
  memptr[f]=mem+1024*count;
  memptre[f]=meme+1024*count;
  count++;
  if (count>=16) count=0;
  }

/* RAM setup */
count=0;
for(f=16;f<128;f++)
  {
  memattr[f]=1;
  memptr[f]=mem+1024*(16+count);
  memptre[f]=meme+1024*(16+count);
  count++;
  }

  mem[0x05c4] = 0x76; /* no edit key */

  mem[0x0003] = 0xfe; /* more space */
  mem[0x0004] = 0xff; /* more space */

//  mem[0x0207] = 0xc9; /* avoid slow/fast/display */
  mem[0x02ba] = 0x76; /* a real halt instead of caused by bit 7 */

  mem[0x0738] = 0x00; /* control e-ppc */
  mem[0x0739] = 0x00;
  mem[0x073a] = 0x00;

  mem[0x347]=0xeb; /* for load */
  mem[0x348]=0xed;
  mem[0x349]=0xfc;
  mem[0x34a]=0xc9;

  mem[0x2fc]=0xed; /* for save */
  mem[0x2fd]=0xfd;
  mem[0x2fe]=0xc9;

  mem[0x02bb] = 0xed; /* for keyboard */
  mem[0x02bc] = 0xf6;
  mem[0x02bd] = 0xc9;

  mem[0x0fa1] = 0xed; /* for inkey$ */
  mem[0x0fa2] = 0xfa;
  mem[0x0fa3] = 0xc3;
  mem[0x0fa4] = 0xed;
  mem[0x0fa5] = 0x0f;

  mem[0x0ee9] = 0xed; /* for input */
  mem[0x0eea] = 0xfb;
  mem[0x0eeb] = 0xc9;

  mem[0x0876] = 0xed; /* for lprint */
  mem[0x0877] = 0xfe;
  mem[0x0878] = 0xc3;
  mem[0x0879] = 0xe2;
  mem[0x087a] = 0x08;

  mem[0x0869] = 0xed; /* for copy */
  mem[0x086a] = 0xf5;
  mem[0x086b] = 0xc9;

  mem[0x10000] = 0x1e; /* font; see sixel_copy() */

/* for str$ patch:
	ld hl,(0x4001)
	push hl
	res 1,(iy+1)
	call 0x15db
	pop hl
	ld (0x4001),hl
	ret
*/
  mem[0x10004] = 0x2a;
  mem[0x10005] = 0x01;
  mem[0x10006] = 0x40;
  mem[0x10007] = 0xe5;
  mem[0x10008] = 0xfd;
  mem[0x10009] = 0xcb;
  mem[0x1000a] = 0x01;
  mem[0x1000b] = 0x8e;
  mem[0x1000c] = 0xcd;
  mem[0x1000d] = 0xdb;
  mem[0x1000e] = 0x15;
  mem[0x1000f] = 0xe1;
  mem[0x10010] = 0x22;
  mem[0x10011] = 0x01;
  mem[0x10012] = 0x40;
  mem[0x10013] = 0xc9;

  mem[0x1bee] = 0x04; /* call the patch */
  mem[0x1bef] = 0x00;
  meme[0x1bef] = 0x01;

  mem[0x16bc] = 0xfb; /* lprint zeros bug fix */

  mem[0x10014] = 0xed; /* for remark */
  mem[0x10015] = 0xff;
  mem[0x10016] = 0xc9;

  mem[0x0c75] = 0x14;
  mem[0x0c76] = 0x00;
  meme[0x0c76] = 0x01;

  mem[0x0f35] = 0xed; /* for pause */
  mem[0x0f36] = 0xf9;
  mem[0x0f37] = 0xc9;

  mem[0x0c0e] = 0xed; /* for scroll */
  mem[0x0c0f] = 0xf8;
  mem[0x0c10] = 0xc9;

  mem[0x1001a] = 0x21; /* for ASCII output to stdout */
  mem[0x1001b] = 0x21; /* originally not used */
  mem[0x1001c] = 0x40;
  mem[0x1001d] = 0x7e;
  mem[0x1001e] = 0xd3;
  mem[0x1001f] = 0xf0; /* the port */
  mem[0x10020] = 0xc9;
}

#ifdef HAVE_SIGNAL
void intHandler(int sig) {
    ctrlcd = 1;
}

void sigHandler(int sig) {
  if (checkscr()) { if (reroute>0) sixel_copy(); else screen(); }
  if (checkhrscr()) hires();
   if (sig==SIGALRM && !ctrlcd) {
     put_keyboard();
     alarm(1);
   }
}
#endif

void list(int lineno, const int mode)
{
	int clineno;

        if (mode==0) z80.run_ret(0x0a2a);

	if (mode==1)
	  mem[0x4001] |= 0x02;
	else
	  mem[0x4001] &= ~0x02;

	mem[0x4001] |= 0x80;

	if (lineno==0) lineno = (mem[0x4024] << 8) | mem[0x4023];

	clineno = (mem[0x400b] << 8) | mem[0x400a];

	if (clineno < lineno) { /* adjust s-top */
	  lineno = clineno;
	  mem[0x4023] = lineno & 0x00ff;
	  mem[0x4024] = (lineno & 0xff00) >> 8;

	}
	for (;;) {

	  z80.c = lineno & 0x00ff;
	  z80.b = (lineno & 0xff00) >> 8;
	  z80.run_ret(0x0733);

/* BERG is 1 if the current line is not displayed because its number
   is higher than that of any shown; adjust s-top */

	  if (mem[0x401e]) {
	    lineno = clineno;
	    mem[0x4023] = lineno & 0x00ff;
	    mem[0x4024] = (lineno & 0xff00) >> 8;
            if (mode==0) z80.run_ret(0x0a2a);
	  } else break;

	}

}

int onlydigits(const char *s)
{
	for (unsigned int i=0; i<strlen(s); i++) if (!isdigit(s[i])) return 0;

	return 1;
}

void delete_line(const int lineno)
{
}

void run_line(int n, int syntax=0)
{
	unsigned int ptr, ptre, llen;

	z80.l = n & 0x00ff;
	z80.h = (n & 0xff00) >> 8;
	mem[0x4007] = z80.l; /* ppc */
	mem[0x4008] = z80.h;
	z80.run_ret(0x09d8); /* line-addr */
	ptr = (z80.h << 8) + z80.l + 4;
	mem[0x4016] = ptr & 0x00ff; /* ch-add */
	mem[0x4017] = (ptr & 0xff00) >> 8;
	mem[0x4001] |= 0x80; /* clear error */
	if (syntax) mem[0x4001] &= ~0x80; else mem[0x4001] |= 0x80;
	mem[0x402d] &= ~0x20; /* not an input line */
	z80.run_ret(0x0cc1); /* line-run */
        ptre = (mem[0x4017] << 8) | mem[0x4016];
	if (mem[0x4000] == 0xff) {
	  printf("0/%d/0x%04x\n", n, ptre);
	  if (syntax) {
	    llen = ptre-ptr+1;
	    mem[ptr-2] = llen & 0x00ff;
	    mem[ptr-1] = (llen & 0xff00) >> 8;
          }
	} else {
	  printf("%d/%d/0x%04x\n", mem[0x4000]+1, n, ptre);
	}
}

void remfloats(int n)
{
	unsigned int ptr, ptrl, ptre, llen;

	z80.l = n & 0x00ff;
	z80.h = (n & 0xff00) >> 8;
	mem[0x4007] = z80.l; /* ppc */
	mem[0x4008] = z80.h;
	z80.run_ret(0x09d8); /* line-addr */
	ptrl = (z80.h << 8) + z80.l + 2;
	llen = (mem[ptrl+1] << 8) | mem[ptrl];
	ptr = ptrl + 2;
        ptre = ptr + llen;

        if (mem[ptr]==0xea) return;

        while (1) {
	  if (mem[ptr]==0x7e) {
	    z80.h = (ptr & 0xff00) >> 8;
	    z80.l = ptr & 0x00ff;
	    z80.b = 0;
	    z80.c = 6;
	    z80.run_ret(0x0a60);
	    ptre -= 6;
	    llen -= 6;
	    mem[ptrl] = llen & 0x00ff;
	    mem[ptrl+1] = (llen & 0xff00) >> 8;
          } else ptr++;
	  if (ptr==ptre) break;
	}
}

void run_program(const int lineno, const int clr)
{
	int n = 0;
	int ll;
	unsigned int ptr;
	size_t fv;	  

	mem[0x403b] = 0; /* slow is not really available */
	if (clr) z80.run_ret(0x149a); /* clear */
	z80.run_ret(0x0a2a); /* cls */

	if (lineno==0) {
	  n = 0;
	  ptr = 0x407d;
	} else {
	  n = lineno;
	  z80.l = lineno & 0x00ff;
	  z80.h = (lineno & 0xff00) >> 8;
	  z80.run_ret(0x09d8);
	  ptr = (z80.h << 8) | z80.l;
	}

	mem[0x4000] = 0xff;
#ifdef HAVE_SELECT
	initts();
#endif
	ctrlcd = 0;

	mem[0x4018] = mem[0x4019] = 0; /* clear x-ptr */

#ifdef HAVE_SIGNAL
	signal(SIGALRM, sigHandler);
	if (doalarm) alarm(1);
#endif

	while (1) {
	  theclock();
	  if (ctrlcd) {
	    printf("D/%d\n", n);
	    mem[0x402b] = n & 0x00ff;
	    mem[0x402c] = (n & 0xff00) >> 8;
	    break;
	  }
	  if (mem[ptr]==0x76) {
	    if (!doquit) printf("%d/%d\n", 0, n);
	    break;
	  }
	  n = (mem[ptr] << 8) | mem[ptr+1];
	  if (dotrace) printf("executing line %d\n",n);
	  ll = (mem[ptr+3] << 8) | mem[ptr+2];

	  if (mem[ptr+4]==0xee && !doquit) { /* at input */
	    if (reroute>=2) sixel_copy();
	    else if (reroute==1) screen();
	  }

	  mem[0x4007] = mem[ptr+1]; /* ppc */
	  mem[0x4008] = mem[ptr];
	  ptr += 4;
	  mem[0x4016] = (ptr & 0x00ff); /* ch-add */
	  mem[0x4017] = (ptr & 0xff00) >> 8;
	  ptr += ll;
	  mem[0x4029] = (ptr & 0x00ff); /* nxtlin */
	  mem[0x402a] = (ptr & 0xff00) >> 8;

	  if (reroute==0) 
	    mem[0x4001] |= 0x02; /* print -> lprint */
	  else
	    mem[0x4001] &= ~0x02;

	  mem[0x402d] &= ~0x20; /* not an input line */
	  mem[0x4001] |= 0x80; /* run instead of syntax check */

	  z80.run_ret(0x0cc1); /* run */

	  fv = (mem[0x4035] << 8) | mem[0x4034];
	  fv += tstates;
	  mem[0x4034] = fv & 0x00ff; /* frames */
	  mem[0x4035] = (fv & 0xff00) >> 8;

	  mem[0x4018] = mem[0x4019] = 0; /* clear x-ptr */

/* check status */
	  if (mem[0x4000] != 0xff) {
	    if (!doquit) printf("%d/%d\n", mem[0x4000]+1, n);
	    n++;
	    mem[0x402b] = n & 0x00ff;
	    mem[0x402c] = (n & 0xff00) >> 8;
	    if (mem[0x4000]==4) {
	      if (reroute>=2) sixel_copy(); else screen();
	    }
	    break;
	  }
	  ptr = (mem[0x402a] << 8) | mem[0x4029]; /* nxtlin */
	}

#ifdef HAVE_SELECT
	resetts();
#endif
#ifdef HAVE_SIGNAL
	signal(SIGALRM, SIG_IGN);
#endif
}

void help()
{
        fprintf(stderr,"usage: zce <options> <file> <argument>\n");
	fprintf(stderr,"  -a        : set an alarm to show the screen every second unless unchanged\n");
#ifdef HAVE_BMP
	fprintf(stderr,"  -b        : read and display bmp (use -oq to write)\n");
#endif
	fprintf(stderr,"  -c <file> : compile (-v and -r may have to precede -c)\n");
	fprintf(stderr,"  -d <addr> : pointer for hires display file\n");
	fprintf(stderr,"  -f        : do not store floating point numbers in program\n");
	fprintf(stderr,"  -h        : shows the options\n");
	fprintf(stderr,"  -k        : enable in-string tokenization\n");
	fprintf(stderr,"  -l        : list\n");
	fprintf(stderr,"  -m        : set the execution mask to the usual 0x7fff\n");
	fprintf(stderr,"  -n <num>  : select file within tzx file (0 to list; see -z)\n");
	fprintf(stderr,"  -o        : output the screen after running (with -q to file `zx81.six`)\n");
	fprintf(stderr,"  -p <file> : read p-file\n");
	fprintf(stderr,"  -q        : quit after compiling, listing, running\n");
	fprintf(stderr,"  -r <num>  : reroute:\n");
	fprintf(stderr,"      0: print -> lprint and enable case sensitivity\n");
	fprintf(stderr,"      1: ASCII copy of \"tv screen\"\n");
	fprintf(stderr,"      2: output to sixel graphics (default)\n");
	fprintf(stderr,"      3: output to html (in combination with -o)\n");
	fprintf(stderr,"  -s <file> : save program as p-file\n");
	fprintf(stderr,"  -t        : enable line number tracing\n");
	fprintf(stderr,"  -u        : unique line numbers\n");
	fprintf(stderr,"  -v        : increase verbosity\n");
	fprintf(stderr,"  -x        : execute program after loading\n");
	fprintf(stderr,"  -z <file> : list tzx file (load when preceded with -n)\n");
	fprintf(stderr,"  -B        : write hex-only representation of ZX's memory (should precede -W)\n");
	fprintf(stderr,"  -H        : hexify everything after REM statements (should precede -W)\n");
#ifdef HAVE_MXML
	fprintf(stderr,"  -M <num>  : wrap margin for writing xml file (should precede -W)\n");
	fprintf(stderr,"  -R <file> : read xml file\n");
	fprintf(stderr,"  -W <file> : write xml file\n");
#endif
        fprintf(stderr,"  without any options but with a file, the defaults are -r0 -qxc <file>\n");
        fprintf(stderr,"  to autorun with sixels, use -xc <file>\n");
}

#ifdef HAVE_READLINE
int put_edit_line()
{
	rl_insert_text(ossbuf.str().c_str());
	rl_redisplay();
	ossbuf.str("");

	return 0;
}

void edit_line(const int id)
{
        z80.l = id & 0x00ff;
	z80.h = (id & 0xff00) >> 8;
	z80.run_ret(0x09d8);
	if (z80.f&0x40) {
	  mem[0x400a] = id & 0x00ff;
	  mem[0x400b] = (id & 0xff00) >> 8;
	  mem[0x4001] |= 0x02;
	  possbuf = 0;
	  z80.d = 0;
	  z80.run_ret(0x075e);
	  possbuf = 1;
	  rl_pre_input_hook = put_edit_line;
	} else fprintf(stderr,"zce: line does not exist -- %d\n", id);
}
#endif

void test()
{
  z80.a = 0xe1;
  z80.run_ret(0x0975);
}

void teste()
{
  mem[0x2000] = 0xcd;
  mem[0x2001] = 0x00;
  mem[0x2002] = 0x00;
  meme[0x2002] = 0x01; /* call at upper bank */
  mem[0x2003] = 0xc9;

  mem[0x10000] = 0xcd;
  mem[0x10001] = 0x04;
  mem[0x10002] = 0x00;
  meme[0x10002] = 0x01;  /* call at upper bank */
  mem[0x10003] = 0xc9;   /* returns within upper bank */

  mem[0x10004] = 0x00;
  mem[0x10005] = 0xc9; /* return to normal bank */
}

void setfloats()
{
  mem[0x1047] = 0xcd; /* dec-to-fp (during syntax check to skip the number) */
  mem[0x1048] = 0xd9;
  mem[0x1049] = 0x14;

  mem[0x104a] = 0xcd; /* syntax-z */
  mem[0x104b] = 0xa6;
  mem[0x104c] = 0x0d;

  mem[0x104d] = 0x20; /* skip reset of stkend if line is executed */
  mem[0x104e] = 0x0c;

  mem[0x104f] = 0x2a; /* get stkend */
  mem[0x1050] = 0x1c;
  mem[0x1051] = 0x40;
  mem[0x1052] = 0x01; /* bc = 5 */
  mem[0x1053] = 0x05; 
  mem[0x1054] = 0x00;
  mem[0x1055] = 0xa7; /* clear carry */
  mem[0x1056] = 0xed; /* subtract bc from hl */
  mem[0x1057] = 0x42;
  mem[0x1058] = 0x22; /* set stkend */
  mem[0x1059] = 0x1c;
  mem[0x105a] = 0x40;

  mem[0x105b] = 0xc3; /* jump forward */
  mem[0x105c] = 0x83;
  mem[0x105d] = 0x10;
}


int main(int argc, char * const argv[])
{
	FILE *in;
#ifdef HAVE_READLINE
	char *line;
	char *expansion;
	int result;
#else
	char expansion[80];
#endif
	char prompt[80];
	int done = 0;
	unsigned int ptr, nAddr=-1;
	int id;
	zline zl;
	int fout = 0;
	int ch;
	char *rompath;
	size_t fv;
        int iarg=-1;

	sf = stdout;

	rompath = getenv("ZX81ROM");
	if (rompath) {
	  strcpy(prompt,rompath);
	  strcat(prompt,"zx81.rom");
	} else {
          strcpy(prompt,argv[0]);
          rompath = (char *) memrchr(prompt,'/',strlen(prompt));
          if (rompath) {
            rompath[1] = '\0';
	    strcat(prompt,"zx81.rom");
          } else {
	    strcpy(prompt,"zx81.rom");
          }
	}

	in = fopen(prompt,"rb");
	if (!in) {
	  fprintf(stderr, "%s not found\n", prompt);
	  return 1;
	}
	fread(mem,1,0x2000,in);
	fclose(in);

	memcpy((void*)(mem+0x2000),(void*)mem,0x2000);

	initmem();
        memset(memscr, 0, nscr);
//	memset(hrold, 0, 0x1800);

	z80.run(0);
	z80.prepare();
 /* clear printer buffer */
	z80.run_ret(0x08e2);

/* for autorun */
        mem[0x4029] = 0x7d;
        mem[0x402a] = 0x40;

/* for RAND and frames_ce() */
	fv = time(NULL);
	mem[0x4034] = fv & 0x00ff; /* frames */
	mem[0x4035] = (fv & 0xff00) >> 8;	
	srand(fv);

/* some interesting data from the environment */

	setstring('P',getenv("PWD"));
	setstring('T',getenv("TERM"));
	setstring('U',getenv("USER"));

	while ((ch = getopt(argc, argv, "ab:c:d:fhklmn:op:qr:s:tuvxz:BHM:R:W:")) != -1) {
          switch (ch) {
#ifdef HAVE_SIGNAL
	    case 'a': doalarm = 1;
		      break;
#endif
#ifdef HAVE_BMP
	    case 'b': readbmp(optarg);
		      break;
#endif
	    case 'c': docompile = 1;
                      compile(optarg);
		      break;
	    case 'd': hrptr = atoih(optarg);
		      break;
	    case 'f': sfloats = 0;
		      setfloats();
		      break;
	    case 'h': help();
	              return 0;
	    case 'l': dolist = 1;
		      break;
	    case 'm': pcmask = 0x17fff; /* for compatibility */
		      break;
	    case 'k': instrtok = 1;
		      break;
	    case 'n': fno = atoi(optarg);
		      break;
	    case 'o': dooutput = 1;
		      break;
	    case 'p': load(optarg);
		      break;
	    case 'q': doquit = 1;
		      break;
	    case 'r': reroute = atoi(optarg);
		      break;
	    case 's': save(optarg);
		      return 0;
	    case 't': dotrace = 1;
		      break;
	    case 'u': uniquelines = 1;
		      break;
	    case 'v': verbose++;
		      break;
	    case 'x': dorun = 1;
		      break;
	    case 'z': loadtzx(optarg,fno);
		      if (fno==0) doquit = 1;
		      break;
	    case 'B': xml_binary=1;
		      break;
	    case 'H': xml_remtok=0;
		      break;
	    case 'M': xml_wrap_margin = atoi(optarg);
		      break;
	    case 'R': rxml(optarg);
		      break;
	    case 'W': wxml(optarg);
		      break;
	    case '?': fprintf(stderr,"%s: use '-h' for help\n",argv[0]);
            default :;
          }
 	}
        argc -= optind;
        argv += optind;

        if (argc>=1 && !docompile) {
          compile(argv[0]);
          reroute = 0;
          dorun = 1;
          doquit = 1;
          if (argc>=2) iarg=1;
        } else if (argc>=1) iarg=0;

	if (iarg>=0) {
	  if (instrtok) {
	    zl.set(argv[iarg]);
	    zl.compile(4);
	    zl.setstring('A');
          } else {
	    setstring('A',argv[iarg]);
	  }
        } else {
          setstring('A',"");
        }

	mem[0x403b] = 0; /* slow is not really available */

#ifdef HAVE_SIGNAL
	signal(SIGINT, intHandler);
	signal(SIGUSR1, sigHandler);
#endif

	if (dolist) list(1,1);
	if (dorun) {
          ptr = (mem[0x402a]<<8) | mem[0x4029];
          if (mem[ptr]!=0x76) {
            if (verbose) printf("starting at address %d\n",ptr);
            ptr = (mem[ptr]<<8) | mem[ptr+1];
            if (verbose) printf("starting at line %d\n",ptr);
            run_program(ptr,0);
          }
        }
	if (dooutput) {	
          if (reroute==3) {
	    hcopy();
	  } else if (reroute==2) {
	    if (doquit) sf = fopen("zx81.six","w");
	    sixel_copy();
#ifdef HAVE_BMP
	    writebmp("zx81.bmp");
#endif
	  } else {
	    screen();
	  }
        }
	if (doquit) {
          return mem[0x4000]+1;
        }

	if (checkscr()) { if (reroute==2) sixel_copy(); else if (reroute==1) screen(); }

#ifdef HAVE_READLINE
	using_history();
	read_history("zce.his");
	rl_pre_input_hook = NULL;
	line = (char *)NULL;
#endif
	if (!verbose) sprintf(prompt, "K> ");

	while (!done) {
	  ctrlcd = 0;
	  frames_ce();	  
	  if (verbose) sprintf(prompt, "[%lu]K> ", tstates);
#ifdef HAVE_READLINE
	  line = readline(prompt);
	  if (!line) exit(1);
	  if (*line) {
	    result = history_expand(line, &expansion);
	    if (result==0 || result==1) {
              add_history(expansion);
#else
	  printf("%s",prompt);
	  gets(expansion);
#endif
	  id = atoi(expansion);

	      if (onlydigits(expansion)) {
		id = atoi(expansion);
		z80.l = id & 0x00ff;
		z80.h = (id & 0xff00) >> 8;
		z80.run_ret(0x09d8);
	  	if (z80.f&0x40) {
	          mem[0x400a] = id & 0x00ff;
	          mem[0x400b] = (id & 0xff00) >> 8;
		  z80.run_ret(0x09f2);
		  z80.run_ret(0x0a60);
		  list(0,0);
	          if (checkscr()) { if (reroute==2) sixel_copy(); else if (reroute==1) screen(); }
		} else fprintf(stderr,"zce: line does not exist -- %d\n", id);
#ifdef HAVE_READLINE
	      } else if (!strncasecmp(expansion,"edit",4)) {
		id = atoi(expansion+4);
		if (id<0) id = 0;
	        if (id==0) id = (mem[0x400b] << 8) | mem[0x400a];
		edit_line(id);
#endif
	      } else if (!strncasecmp(expansion,"run",3)) {
	        id = atoi(expansion+3);
		run_program(id,1);
	      } else if (!strncasecmp(expansion,"goto",4)) {
	        id = atoi(expansion+4);
		run_program(id,0);
	      } else if (!strcasecmp(expansion,"clear")) {
		z80.run_ret(0x149a);
	      } else if (!strcasecmp(expansion,"cls")) {
		z80.run_ret(0x0a2a);
	      } else if (!strcasecmp(expansion,"cont")) {
	        id = (mem[0x402c] << 8) | mem[0x402b];
		run_program(id,0);
	      } else if (!strcasecmp(expansion,"copy")) {
		screen();
	      } else if (!strcasecmp(expansion,"scopy")) {
		sixel_copy();
	      } else if (!strcasecmp(expansion,"hcopy")) {
		hcopy();
	      } else if (!strncasecmp(expansion,"list",4)) {
		z80.run_ret(0x0a2a);
	        id = atoi(expansion+4);
	        if (id>0) {
	          mem[0x400a] = id & 0x00ff;
	          mem[0x400b] = (id & 0xff00) >> 8;
	        }
		list(id,0);
//	        if (reroute<=1) screen(); else if (reroute>=2) sixel_copy();
	      } else if (!strncasecmp(expansion,"llist",5)) {
	        id = atoi(expansion+5);
		list(id,1);
	      } else if (!strcasecmp(expansion,"new")) {
		id = (mem[0x4005] << 8) | mem[0x4004];
		id--;
		mem[0x0003] = id & 0x00ff; /* ramtop functionality */
		mem[0x0004] = (id & 0xff00) >> 8;
		z80.run(0);
		z80.run_ret(0x08e2); /* clear printer buffer */
	      } else if (!strcasecmp(expansion,"svlist")) {
	        cns.set_ptr(0x4000);
		cns.system_variables();
	      } else if (!strcasecmp(expansion,"vlist")) {
		cns.vlist();
	      } else if (!strcasecmp(expansion,"svars")) {
		cns.save_vars();
	      } else if (!strcasecmp(expansion,"mvars")) {
		cns.merge_vars();
	      } else if (!strcasecmp(expansion,"sscreen")) {
		cns.save_screen();
	      } else if (!strcasecmp(expansion,"lscreen")) {
		cns.load_screen();
	      } else if (id>0 || (expansion[0]>='A' && expansion[0]<='Z')) {
#ifdef HAVE_SELECT
                if (id==0) if (!strncasecmp(expansion,"input",5)) initts();
#endif
	        zl.set(expansion);
		zl.compile(id>0?0:2);
#ifdef HAVE_READLINE
		if (zl.insert(1)) {
		  if (id>0) edit_line(id);
		} else if (id>0) {
	          mem[0x400a] = id & 0x00ff;
	          mem[0x400b] = (id & 0xff00) >> 8;
		  list(0,0);
//	          if (reroute<=1) screen(); else if (reroute>=2) sixel_copy();
		}
#else
		zl.insert();
#endif
#ifdef HAVE_SELECT
                if (id==0) if (!strncasecmp(expansion,"input",5)) resetts();
#endif
	      } else
	      switch (expansion[0]) {
	        case '#' : break;
	        case 'w' : ptr = atoih(expansion+1);
		           printf("%02X%02X\n",mem[ptr+1],mem[ptr]);
	                   break;
	        case 'b' : if (expansion[1]!='\0') {
			     ptr = atoih(expansion+1);
			     z80.breakpoint(ptr);
			   } else z80.cancel();
			   break;
	        case 'c' : if (expansion[1]!='\0') {
			     z80.run_ret(atoih(expansion+1));
			   } else if (mem[z80.pc]==0xcd) {
			     ptr = z80.pc;
			     z80.run_ret((mem[z80.pc+2]<<8)|mem[z80.pc+1]);
			     z80.pc = ptr+3;
			   }
#ifdef DISASSEMBLER
			   disassemble(z80.pc);
			   z80.show(0);
#else
			   z80.show(1);
#endif
			   break;
#ifdef DISASSEMBLER
	        case 'd' : if (expansion[1]!='\0') nAddr = z80.pc = atoih(expansion+1);
			   nAddr = disassemble(nAddr);
			   break;
#endif
		case 'f' : ptr = atoih(expansion+1);
			   z80.a = mem[ptr];
			   z80.e = mem[ptr+1];
			   z80.d = mem[ptr+2];
			   z80.c = mem[ptr+3];
			   z80.b = mem[ptr+4];
			   z80.run_ret(0x158d);
			   z80.show();
		case 'g' : run_line(atoi(expansion+1));
			   break;
		case 'a' : run_line(atoi(expansion+1),1); /* add floats */
			   break;
		case 'y' : remfloats(atoi(expansion+1)); /* yank floats */
			   break;
		case 'h' : hires();
			   break;
		case 'j' : if (expansion[1]!='\0') z80.run(atoih(expansion+1)); else z80.run(z80.pc);
#ifdef DISASSEMBLER
			   disassemble(z80.pc);
			   z80.show(0);
#else
			   z80.show(1);
#endif
			   break;
		case 'k' : id = atoih(expansion+1);
			   z80.c = id & 0x00ff;
			   z80.b = (id & 0xff00) >> 8;
			   z80.run_ret(0x07bd);
			   id = mem[(z80.h<<8)|z80.l];
			   printf("%02X %d %c\n", id, id, zx2ascii[id]);
			   break;
		case 'n' : id = atoi(expansion+1);
			   z80.l = id & 0x00ff;
			   z80.h = (id & 0xff00) >> 8;
			   z80.run_ret(0x09d8);
			   z80.show();
			   break;
	        case 'o' : reroute = atoi(expansion+1);
			   break;
	        case 'p' : id = mem[atoih(expansion+1)]; /* peek */
			   printf("%02X %c\n",id,zx2ascii[id&0x3f]);
			   break;
	        case 'e' : id = meme[atoih(expansion+1)]; /* peek */
			   printf("%02X %c\n",id,zx2ascii[id&0x3f]);
			   break;
	        case 'q' : done = 1;
			   break;
	        case 'r' : switch (expansion[1]) {
			     case 'a' : z80.a = atoih(expansion+2); break;
			     case 'f' : z80.f = atoih(expansion+2); break;
			     case 'b' : z80.b = atoih(expansion+2); break;
			     case 'c' : z80.c = atoih(expansion+2); break;
			     case 'd' : z80.d = atoih(expansion+2); break;
			     case 'e' : z80.e = atoih(expansion+2); break;
			     case 'h' : z80.h = atoih(expansion+2); break;
			     case 'l' : z80.l = atoih(expansion+2); break;
			     case 'x' : z80.ix = atoih(expansion+2); break;
			     case 'y' : z80.iy = atoih(expansion+2); break;
			     case 's' : z80.sp = atoih(expansion+2); break;
			     default  : z80.zero(); tstates=0;
	                   }
			   break;
	        case 's' : if (expansion[1]!='\0') {
			      z80.pc = atoih(expansion+1);
#ifdef DISASSEMBLER
			      disassemble(z80.pc);
			      printf("\n");
#endif
			   }
			   z80.step(0);
#ifdef DISASSEMBLER
			   disassemble(z80.pc);
			   z80.show(0);
#else
			   z80.show(1);
#endif
			   break;
		case 't' : fout = !fout;
			   if (fout) {
			     sf = fopen("zx81.six","w");
			   } else {
			     fclose(sf);
			     sf = stdout;
			   }
			   break;
		case 'u' : test(); break;
	        case 'v' : printf("verbose: %d\n",z80.toggle_verbose());
			   break;
	        case 'x' : id = atoih(expansion+1);
			   printf("%04x %d\n", id, id);
			   break;
	        case 'z' : z80.show();
			   break;
	        default  : fprintf(stderr,"zce: no such command -- %s\n", expansion);
	      }
#ifdef HAVE_READLINE
	    } else {
	      printf("%s\n", expansion);
	    }
	    free(line);
	    free(expansion);
	    if (checkscr()) { if (reroute==2) sixel_copy(); else if (reroute==1) screen(); }
 	  }
#endif
	}

	if (fout) fclose(sf);
	if (ef) fclose(ef);

#ifdef HAVE_READLINE
	write_history("zce.his");
#endif

	return mem[0x4000]+1;
}
