#include <stdio.h>
#include "zline.h"
#include "z80.h"
#include "zacvt.h"

extern unsigned char mem[MSIZE];
extern z80_processor z80;
extern int verbose, reroute, instrtok;
int htoi(const char *s, int &v);

// #define DEBUG

/* converts a token (a sequence of letters) to the abbreviated byte code */

int zline::token()
{
	int i, n, ntok;
	unsigned int ptr;
	unsigned char c;

#ifdef DEBUG
	printf("[");
	for (i=0; i<tok.length(); i++) printf("%c",zx2ascii[tok[i]]);
	printf("]\n");
#endif
	n = tok.length()-1;

	c = tok[n] | 0x80;
	ptr = 0x01fb;
	ntok = 0;

	while (1) {
	  if (ptr<0x0111) return 0;
	  if (mem[ptr]&0x80) ntok++;
	  if (c==mem[ptr]) {
	    for (i=1; i<=n; i++) {
	      if (mem[ptr-i]!=(tok[n-i]&0x7f)) break;
	    }
	    if ((i>n) && (mem[ptr-i]&0x80)) {
#ifdef DEBUG
	      printf("%04x\n",ptr);
#endif
	      if (ntok <= 3) return 67-ntok; else return 259-ntok;
	    } else ptr--;
	  } else {
	    ptr--;
	  }
	}

	return 0;
}

void zline::compile(int linf)
{
	unsigned i, j, ic=0, tokf, tokn=0, tokparse=1, cmd=0;
	static int lln=10;
	unsigned int instring=0;
        int value;

	lineno = 0;

	if (verbose)
	  printf("compiling: %s\n", s.c_str());

	tokf = i = 0;

	while (1) {

	  if (linf>=2) {

#ifdef DEBUG
printf("%d/%d %c/\n",tokf,i,s[i]);
#endif

	    if (linf==2 && s[i]!=' ') {
	      ic = i;
	      linf++;
	    }

	    if (instrtok) {

	    if (s[i]=='\"') {
	       if (instring && s[i+1]=='\"') {
	         s[i] = (char) 0xc0;
		 i++;
		 s.erase(i,1);
		 if (s[i]=='\0') break; /* syntax error */
		 continue;
	       } else {
		 if (cmd!=0xea) {
		   instring = !instring;
	         }
	       }
	    }

	    } else {

	    if (s[i]=='\"') {
	       if (!tokparse && s[i+1]=='\"') {
	         s[i] = (char) 0xc0;
		 i++;
		 s.erase(i,1);
		 if (s[i]=='\0') break; /* syntax error */
		 continue;
	       } else {
		 if (cmd!=0xea) {
		   if (tokparse>0) tokparse--; else tokparse=1;
	         }
	       }
	    } else if (tokparse==2) tokparse=1;

	    }

#ifdef DEBUG
	    printf("tokparse: %d\n",tokparse);
#endif
	    if (tokparse) {

	    if ((s[i]>='A' && s[i]<='Z') || (s[i]>='a' && s[i]<='z')) {
	      if (!tokf) j = i;
	      tokf = 1;
	      if (reroute==0 && instring)
	        s[i] = ascii2zxcs[(int)s[i]];
	      else
	        s[i] = ascii2zx[(int)s[i]];
	    } else {
	      tokn = 0;
	      if (tokf) {
	        if (s[i]=='$') {
	          s[i] = ascii2zx[(int)s[i]];
		  i++;
		}
		tok = s.substr(j,i-j);
		tokn = token();
#ifdef DEBUG
		printf("tokn: %x\n",tokn);
#endif
	        if (tokn>0) {
	          if (tokn==0xc5) tokparse=2; /* VAL function */
		  if (instrtok && instring && s[i]==' ') i++; /* makes instrtok somewhat better */
#ifdef DEBUG
		  printf("erasing: %d,%d\n",j,i-j-1);
#endif
		  s.erase(j,i-j-1);
		  s[j] = (char) tokn;
		  i = j;
		  if (linf==3) {
		    cmd = tokn;
		    if (cmd==0xea) tokparse=0;
		    ic = i;
		    linf++;
		  }
		}
	      }

	      if (s[i]=='*' && s[i+1]=='*') {
	        s[i] = (char) 0xd8;
		i++;
		s.erase(i,1);
#ifdef DEBUG
		printf("erasing: %d\n",i);
#endif
	        continue;
	      } else  if ((s[i]=='<' && s[i+1]=='=') ||
	          (s[i]=='>' && s[i+1]=='=') ||
	          (s[i]=='<' && s[i+1]=='>')) {
	        j = i+2;
	        s[i] = ascii2zx[(int)s[i]];
	        s[i+1] = ascii2zx[(int)s[i+1]];
		tok = s.substr(i,j-i);
	        tokn = token();
#ifdef DEBUG
		printf("%x\n",tokn);
#endif
		s[i] = char(tokn);
		s.erase(i+1,1);
	        tokf = 0;
	      }

	      if (s[i]=='\0') break;
	      if (!tokn) s[i] = ascii2zx[(int)s[i]];
	      tokf = 0;

	    } /* A-Z */

	    if (!instring && s[i]=='\0') {s.erase(i,1); i--;}

	    } else { /* tokparse */

	      if (s[i]==' ' && i==(ic+1)) s.erase(i,1);
	      if (s[i]=='\0') break;
	      if (s[i]=='#') {
		htoi(s.substr(i+1,2).c_str(),value);
		if (value>=0) {
		  s.erase(i,2);
		  s[i] = (unsigned char) value;
		} else s[i] = ascii2zx[(int)s[i]];
	      } else {
	        if (reroute==0)
	          s[i] = ascii2zxcs[(int)s[i]];
	        else
	          s[i] = ascii2zx[(int)s[i]];
	      }

	    } /* tokparse */

	  } else { /* linf */

	    if (s[i]>='0' && s[i]<='9') {
	      linf = 1;
	      lineno = 10*lineno + s[i]-48;
	    } else if (s[i]!=' ') {
	      if (linf==0) {
		lineno = lln;
	      } else lln = lineno;
	      linf = 2;
	      i--;
#ifdef DEBUG
	      printf("LINE:%d\n", lineno);
#endif
	    }

	  } /* linf */

	  i++;

	} // while

	s = s.substr(ic);
}

#ifdef DEBUG
#undef DEBUG
#endif

int zline::insert(int unique)
{
	int i, n, nn;
	unsigned int ptr, sptr;
	unsigned char c;

	n = s.length();

#ifdef DEBUG
	printf("Inserting:\n");
	for (i=0; i<n; i++) printf("%d %x %c\n", i, (int)(s[i]&0xff), zx2ascii[s[i]&0x3f]);
#endif

// #undef DEBUG

	if (lineno > 0) {

	  z80.l = lineno & 0x00ff;
	  z80.h = (lineno & 0xff00) >> 8;
	  z80.run_ret(0x09d8); /* line-addr */

/* if the line exists, place the new line before a line with a higher number, or replace it */

	  if (z80.f&0x40) {
	    if (unique) {
	      z80.run_ret(0x09f2);
	      z80.run_ret(0x0a60);
	    } else {
	      lineno++;
	      z80.l = lineno & 0x00ff;
	      z80.h = (lineno & 0xff00) >> 8;
	      z80.run_ret(0x09d8); /* line-addr */
	      lineno--;
	    }
	  }

	  ptr = (z80.h << 8) | z80.l;
#ifdef DEBUG
printf("BOL:%x\n", ptr);
#endif
	  nn = n+5;
	  z80.b = (nn & 0xff00) >> 8;
	  z80.c = nn & 0x00ff;
	  ptr--;
	  z80.h = (ptr & 0xff00) >> 8;
	  z80.l = (ptr & 0x00ff);
	  z80.run_ret(0x099e); /* make-room */
	  ptr++;

	  sptr = ptr+4;
	  mem[ptr++] = (lineno & 0xff00) >> 8;
	  mem[ptr++] = lineno & 0x00ff;
	  nn = n+1;
	  mem[ptr++] = nn & 0x00ff;
	  mem[ptr++] = (nn & 0xff00) >> 8;
	  for (i=0; i<n; i++) mem[ptr++] = (s[i]&0xff);
	  mem[ptr] = 0x76;
#ifdef DEBUG
printf("EOL:%x\n", ptr);
#endif

	} else {
	  nn = n+1;
	  z80.b = (nn & 0xff00) >> 8;
	  z80.c = (nn & 0x00ff);
	  z80.run_ret(0x0030); /* bc-spaces */
//	  z80.show();
	  ptr = (z80.d << 8) | z80.e;
	  sptr = ptr;
	  for (i=0; i<n; i++) mem[ptr++] = (s[i]&0xff);
	  mem[ptr] = 0x76;
	}

// syntax check, also for adding floating point numbers

	z80.run_ret(0x14bc); /* set-mem */
	mem[0x4000] = 0xff; /* clear error */

	mem[0x4001] &= ~0x80; /* syntax check */
	mem[0x4016] = sptr & 0x00ff;
	mem[0x4017] = (sptr & 0xff00) >> 8;
	z80.run_ret(0x0cde); /* run */
	
#ifdef DEBUG
	printf("CH_ADD:%x%x X-PTR:%x%x\n", mem[0x4017],mem[0x4016], mem[0x4019],mem[0x4018]);
#endif

	if (mem[0x4000] != 0xff) {
	  printf("%d/%d\n", mem[0x4000]+1, lineno);
	  mem[0x4001] |= 0x02;
	  z80.h = (sptr & 0xff00) >> 8;
	  z80.l = sptr & 0x00ff;
	  z80.run_ret(0x0766); /* list line */
	  z80.run_ret(0x14a3); /* x-temp */
	  if (lineno==0) return 1;
	}

#ifdef DEBUG
	printf("CH_ADD:%x%x X-PTR:%x%x\n", mem[0x4017],mem[0x4016], mem[0x4019],mem[0x4018]);
#endif

// adjust length

	if (lineno > 0) {

	n++;
	while (1) {
	  c = mem[ptr];
	  if (c==0x7e) { n+=5; ptr+=5; }
	  if (c==0x76) break;
	  n++;
	  ptr++;
	}
	mem[sptr-2] = n & 0x00ff;
	mem[sptr-1] = (n & 0xff00) >> 8;

	if (mem[0x4000] != 0xff) return 1;

	} else {

	z80.run_ret(0x14bc); /* set-mem */
	mem[0x4000] = 0xff;

	mem[0x4001] |= 0x82; /* run to stdout */
	mem[0x4016] = sptr & 0x00ff; /* ch-add */
	mem[0x4017] = (sptr & 0xff00) >> 8;
	z80.run_ret(0x0cde); /* run */
	z80.run_ret(0x14a3); /* x-temp */
	if (mem[0x4000] != 0xff) {
	  printf("%d/%d\n", mem[0x4000]+1,lineno);
	  return 1;
	}

	}

	return 0;
}

void zline::setstring(unsigned char vr)
{
	unsigned int ptr;
	int i, n, ns, nc;

	ns = s.length();

#ifdef DEBUG
	printf("Inserting:\n");
	for (i=0; i<ns; i++) printf("%d %x %c\n", i, (int)(s[i]&0xff), zx2ascii[s[i]&0x3f]);
#endif
	vr = ascii2zx[vr];

	mem[0xfff4] = vr;
	mem[0xfff5] = ascii2zx['$'];
	mem[0xfff6] = 0x76;

	mem[0x4001] |= 0x80; /* run */

	mem[0x4016] = 0xf4; /* ch-add */
	mem[0x4017] = 0xff;

	z80.run_ret(0x111c); /* look-vars */

	if (z80.f&0x01) {
	  z80.h = mem[0x4011];
	  z80.l = mem[0x4010];
	  nc = ns + 3;
	  z80.b = (nc & 0xff00) >> 8;
	  z80.c = nc & 0x00ff;
	  z80.run_ret(0x99e);
	  ptr = (mem[0x4011]<<8) | mem[0x4010];
	} else {
	  ptr = (z80.h<<8) | z80.l;
	  n = (mem[ptr+2]<<8) | mem[ptr+1];
	  n = ns - n;
	  z80.b = (n & 0xff00) >> 8;
	  z80.c = n & 0x00ff;
	  if (n>0) {
	    z80.run_ret(0x099e);
	    z80.b = (n & 0xff00) >> 8;
	    z80.c = n & 0x00ff;
	  } else if (n<0) {
	    n = -n;
	    z80.b = (n & 0xff00) >> 8;
	    z80.c = n & 0x00ff;
	    z80.run_ret(0x0a60);
	  }
	}
	mem[ptr++] = (vr & 0x5f) | 0x40;
	mem[ptr++] = ns & 0x00ff;
	mem[ptr++] = (ns & 0xff00) >> 8;
	for (i=0; i<ns; i++) mem[ptr++] = s[i];
}

