#include <string>
#include <mxml.h>
#include "z80.h"
#include "zacvt.h"
#include "zline.h"

using namespace std;

extern z80_processor z80;
extern unsigned char mem[MSIZE];
extern int xml_wrap_margin, xml_remtok, xml_binary;
static unsigned int rptr;

int atoih(const char *s, int &v)
{
        int n;
        n = sscanf(s,"%i",&v);
        return n;
}

void htoi(const char *s, int &v)
{
  v = -1;
  if (s[0]>='0' && s[0]<='9') v = (s[0]-48)*16;
  if (s[0]>='A' && s[0]<='Z') v = (s[0]-55)*16;
  if (s[0]>='a' && s[0]<='z') v = (s[0]-87)*16;
  if (v<0) return;
  if (s[1]>='0' && s[1]<='9') v += (s[1]-48);
  if (s[1]>='A' && s[1]<='Z') v += (s[1]-55);
  if (s[1]>='a' && s[1]<='z') v += (s[1]-87);
}

int tok2str(unsigned char c, string& s)
{
  unsigned int ptr;

  s = "";
  if (c<0x40 || (c>0x42 && c<0xc0)) return 0;
  z80.a = c;
  z80.run_ret(0x975);
  ptr = (z80.b << 8) | z80.c;
  do {
    c = mem[ptr++];
    s += tolower(zx2ascii[c&0x3f]);
  } while (!(c&0x80));

  return 1;
}

int str2tok(string s)
{
	int i, n, ntok;
	unsigned int ptr;
	unsigned char c;

	n = s.length()-1;
	for (i=0; i<=n; i++) s[i] = ascii2zx[(int)s[i]];

	c = s[n] | 0x80;
	ptr = 0x01fb;
	ntok = 0;

	while (1) {
	  if (ptr<0x0111) return 0;
	  if (mem[ptr]&0x80) ntok++;
	  if (c==mem[ptr]) {
	    for (i=1; i<=n; i++) {
	      if (mem[ptr-i]!=(s[n-i]&0x7f)) break;
	    }
	    if ((i>n) && (mem[ptr-i]&0x80)) {
	      if (ntok <= 3) return 67-ntok; else return 259-ntok;
	    } else ptr--;
	  } else {
	    ptr--;
	  }
	}

	return 0;
}

static const char *whitespace_cb(mxml_node_t *node, int where)
{
  const char *name;
  name = mxmlGetElement(node);

  if (where == MXML_WS_AFTER_CLOSE && strcmp(name,"lineno") && strcmp(name,"hex") && strcmp(name,"token") && strcmp(name,"sysvar") && strcmp(name,"var"))
      return ("\n");

  if (!strcmp(name,"sysvars") || !strcmp(name,"vars"))
    if (where == MXML_WS_BEFORE_CLOSE)
      return ("\n");

  if (!strcmp(name,"sysvar") || !strcmp(name,"var") || !strcmp(name,"e_byte"))
    if (where == MXML_WS_BEFORE_OPEN)
      return ("\n\t");

  if (!strcmp(name,"statement") || !strcmp(name,"line"))
    if (where == MXML_WS_BEFORE_OPEN)
      return ("\t");

  if (!strcmp(name,"program") || !strcmp(name,"d_file"))
    if (where==MXML_WS_AFTER_OPEN)
      return ("\n");

  if (!strcmp(name,"sinclair"))
    if (where == MXML_WS_BEFORE_OPEN || where==MXML_WS_AFTER_OPEN)
      return ("\n");

  return (NULL);
}

void wxml_sysvar(mxml_node_t *data, unsigned int &ptr, int n, const char *name)
{
    mxml_node_t *node;
    char buf1[7], buf2[7];

    node = mxmlNewElement(data, "sysvar");
    if (n==1)
      sprintf(buf1,"0x%02x",mem[ptr]);
    else
      sprintf(buf1,"0x%02x%02x",mem[ptr+1],mem[ptr]);
    sprintf(buf2,"0x%04x",ptr);
    mxmlElementSetAttr(node,"addr",buf2);
    mxmlElementSetAttr(node,"name",name);
    if (n==1) mxmlElementSetAttr(node,"bytes","0x01"); else mxmlElementSetAttr(node,"bytes","0x02");
    mxmlElementSetAttr(node,"value",buf1);
    ptr += n;
}

void wxml_sysbuf(mxml_node_t *data, unsigned int &ptr, int n, const char *name)
{
    mxml_node_t *node;
    char buf1[67], buf2[7];
    int i;

    node = mxmlNewElement(data, "sysvar");
    sprintf(buf2,"0x%04x",ptr);
    mxmlElementSetAttr(node,"addr",buf2);
    mxmlElementSetAttr(node,"name",name);
    sprintf(buf2,"0x%02x",n);
    mxmlElementSetAttr(node,"bytes",buf2);
    for (i=0; i<n; i++) sprintf(buf1+i*2,"%02x",mem[ptr++]);
    buf1[n*2] = '\0';
    mxmlElementSetAttr(node,"buffer",buf1);
}

void wxml_sysvars(mxml_node_t *data)
{
  unsigned int ptr;

	ptr = 0x4009;
	wxml_sysvar(data,ptr,1,"VERSN ");
	wxml_sysvar(data,ptr,2,"E_PPC ");
	wxml_sysvar(data,ptr,2,"D_FILE");
	wxml_sysvar(data,ptr,2,"DF-CC ");
	wxml_sysvar(data,ptr,2,"VARS  ");
	wxml_sysvar(data,ptr,2,"DEST  ");
	wxml_sysvar(data,ptr,2,"E-LINE");
	wxml_sysvar(data,ptr,2,"CH-ADD");
	wxml_sysvar(data,ptr,2,"X-PTR ");
	wxml_sysvar(data,ptr,2,"STKBOT");
	wxml_sysvar(data,ptr,2,"STKEND");
	wxml_sysvar(data,ptr,1,"BERG  ");
	wxml_sysvar(data,ptr,2,"MEM   ");
	wxml_sysvar(data,ptr,1,"N/U   ");
	wxml_sysvar(data,ptr,1,"DF-SZ ");
	wxml_sysvar(data,ptr,2,"S-TOP ");
	wxml_sysvar(data,ptr,2,"LAST-K");
	wxml_sysvar(data,ptr,1,"DEBNCE");
	wxml_sysvar(data,ptr,1,"MARGIN");
	wxml_sysvar(data,ptr,2,"NXTLIN");
	wxml_sysvar(data,ptr,2,"OLDPPC");
	wxml_sysvar(data,ptr,1,"FLAGX ");
	wxml_sysvar(data,ptr,2,"STRLEN");
	wxml_sysvar(data,ptr,2,"T-ADDR");
	wxml_sysvar(data,ptr,2,"SEED  ");
	wxml_sysvar(data,ptr,2,"FRAMES");
	wxml_sysvar(data,ptr,2,"COORDS");
	wxml_sysvar(data,ptr,1,"PR-CC ");
	wxml_sysvar(data,ptr,1,"S-POSN");
	wxml_sysvar(data,ptr,1,"S-POSN");
	wxml_sysvar(data,ptr,1,"CDFLAG");
	wxml_sysbuf(data,ptr,33,"PRBUFF");
	wxml_sysbuf(data,ptr,30,"MEMBOT");
	wxml_sysvar(data,ptr,1,"N/U   ");
}

void wxml_program(mxml_node_t *data)
{
  char buf[97];
  unsigned int ptr, ptr2;
  unsigned char c, cc, cmd;
  mxml_node_t *node, *snode, *nnode;
  int lineno, linelen;
  string line, hexline, tok;
  int tokf, hexf, hexx=0;
  int i, j;

  ptr = 0x407d;
  ptr2 = mem[0x400c] + (mem[0x400d] << 8);

  while (ptr<ptr2) {
    node = mxmlNewElement(data, "statement");
    sprintf(buf,"0x%04x",ptr);
    mxmlElementSetAttr(node,"addr",buf);
    linelen = (mem[ptr+3]<<8) | mem[ptr+2];
    sprintf(buf,"0x%04x",linelen);
    mxmlElementSetAttr(node,"bytes",buf);
    nnode = mxmlNewElement(node, "lineno");
    lineno = (mem[ptr]<<8) | mem[ptr+1];
    sprintf(buf,"%d",lineno);
    mxmlNewText(nnode,0,buf);
    line = "";
    ptr += 4;
    cmd = 0;
    for (j=1; j<=linelen; j++) {
      c = mem[ptr++];
      hexf = (((c>=0x01 && c<=0x0a) || (c>=0x43 && c<=0xa5) || c==0xc3) && c!=0x7e) || (cmd==0xea && !xml_remtok);
      if (!hexf && hexx) {
        hexx = 0;
        mxmlNewText(snode, 0, hexline.c_str());
      }
      tokf = tok2str(c,tok);
      if (!cmd) cmd = c;
      if (hexf) {
        if (!hexx) {
	  hexx = 1;
          mxmlNewText(node, 0, line.c_str());
	  line = "";
          snode = mxmlNewElement(node, "hex");
          hexline = "";
        }
	sprintf(buf,"%02x",c);
        hexline += buf;
      } else if (c==0x7e && cmd!=0xea) {
	ptr--;
        j+=5;
        mxmlNewText(node, 0, line.c_str());
	line = "";
        snode = mxmlNewElement(node, "hex");
	hexline = "";
        for (i=0; i<=5; i++) {
	  sprintf(buf,"%02x",mem[ptr++]);
	  hexline += buf;
	}
        mxmlNewText(snode, 0, hexline.c_str());
      } else if (tokf) {
        mxmlNewText(node, 0, line.c_str());
	line = "";
        snode = mxmlNewElement(node, "token");
        mxmlNewText(snode, 0, tok.c_str());
      } else {
	cc = zx2ascii[c&0x3f];
	if (!(c&0x80)) cc = tolower(cc);
        line += cc;
      }
    }
    mxmlNewText(node, 0, line.c_str());
  }
}

void wxml_d_file(mxml_node_t *data)
{
  unsigned char buf[67];
  unsigned int ptr, ptr1, ptr2;
  int j=0, lineno=0;
  mxml_node_t *node, *snode;

  ptr1 = mem[0x400c] + (mem[0x400d] << 8);
  ptr2 = mem[0x4010] + (mem[0x4011] << 8);

  for (ptr=ptr1; ptr<ptr2-1; ) {
    if (mem[ptr]==0x76 || j==0) {
      node = mxmlNewElement(data, "line");
      sprintf((char*)buf,"0x%04x",ptr);
      mxmlElementSetAttr(node,"addr",(char*)buf);
      sprintf((char*)buf,"0x%02x",lineno);
      mxmlElementSetAttr(node,"lineno",(char*)buf);
      snode = mxmlNewElement(node, "hex");
      j = 0;
      lineno++;
      ptr++;
    }
    if (mem[ptr]!=0x76) {
      sprintf((char*)(buf+j*2),"%02x",mem[ptr]);
      if (j<32) j++;
      ptr++;
    }
    if (mem[ptr]==0x76) {
      sprintf((char*)(buf+j*2),"%02x",mem[ptr]);
      if (j<32) j++;
      mxmlNewText(snode, 0, (char*)buf);
      sprintf((char*)buf,"0x%02x",j+1);
      mxmlElementSetAttr(node,"bytes",(char*)buf);
    }
  }
}

void wxml_d_file_compress(mxml_node_t *data)
{
  unsigned char buf[65];
  char bufr[3];
  int c, pc;
  unsigned int ptr;
  int i, j, repeat;
  mxml_node_t *node, *snode;

  ptr = mem[0x400c] + (mem[0x400d] << 8);

  for (i=0; i<24; i++) {
    node = mxmlNewElement(data, "line");
    sprintf((char*)buf,"0x%04x",ptr);
    mxmlElementSetAttr(node,"addr",(char*)buf);
    ptr++;
    pc = -1;
    repeat = j = 0;
    while ((c=mem[ptr])!=0x76) {
      if (repeat==0) {
        snode = mxmlNewElement(node, "hex");
        j = 0;
        repeat = 1;
      }
      if (c!=pc || j==0) {
        sprintf((char*)(buf+j*2),"%02x",c);
        j++;
      } else {
        repeat++;
      }
      if (mem[ptr+1]==0x76 || (mem[ptr+1]!=c && repeat>1)) {
        if (repeat>1) {
          sprintf(bufr,"%d",repeat);
          mxmlElementSetAttr(snode,"n",bufr);
        }
        mxmlNewText(snode, 0, (char*)buf);
        repeat = 0;
      }
      ptr++;
      pc = c;
    }
  }
}

void wxml_vars(mxml_node_t *data)
{
	unsigned int ptr, ptr1, ptr2;
	unsigned char c, mask;
	string s;
	char buf[7];
	mxml_node_t *node, *snode;

	z80.h = mem[0x4011]; /* vars */
	z80.l = mem[0x4010];
	while (1) {
	  ptr = ptr2 = (z80.h << 8) + z80.l;
	  c = mem[ptr];
	  if (c==0x80) break;
          node = mxmlNewElement(data, "var");
          sprintf(buf,"0x%04x",ptr);
          mxmlElementSetAttr(node,"addr",buf);
	  mask = c & 0xe0;
	  switch (mask) {
	    case 0x60 : s = zx2ascii[c&0x3f];
			break;
	    case 0xa0 : s = zx2ascii[c&0x3f];
			while (1) {
		          c = mem[++ptr];
			  s += zx2ascii[c&0x3f];
			  if ((c&0xc0)==0x80) break;
			}
			break;
	    case 0x40 : s = zx2ascii[(c&0x3f)+0x20]; s+='$';
			break;
	    case 0x80 : s = zx2ascii[(c&0x3f)+0x20];
			break;
	    case 0xc0 : s = zx2ascii[(c&0x3f)+0x20]; s+='$';
			break;
	    case 0xe0 : s = zx2ascii[c&0x3f];
			break;
	    default   :;
	  }
	  mxmlElementSetAttr(node,"name",s.c_str());
	  z80.run_ret(0x09f2);
	  ptr1 = (z80.h << 8) | z80.l;
	  ptr2 = (z80.d << 8) | z80.e;
	  sprintf(buf,"0x%04x",ptr2-ptr1);
	  mxmlElementSetAttr(node,"bytes",buf);
          snode = mxmlNewElement(node, "hex");
	  s = "";
	  for (ptr=ptr1; ptr<ptr2; ptr++) {
	    sprintf(buf,"%02x",mem[ptr]);
	    buf[2] = '\0';
	    s += buf;
	  }
	  mxmlNewText(snode, 0, s.c_str());
	  z80.h = z80.d;
	  z80.l = z80.e;
	}
        node = mxmlNewElement(data, "e_byte");
	sprintf(buf,"0x%04x",ptr2);
	mxmlElementSetAttr(node,"addr",buf);
	sprintf(buf,"0x%02x",mem[ptr2]);
	mxmlElementSetAttr(node,"value",buf);
}

void wxml_hex(mxml_node_t *data)
{
    string line;
    char buf[7];
    unsigned int ptr, ptr1, ptr2;

    ptr1 = 0x4009;
    ptr2 = (mem[0x4015] << 8) | mem[0x4014];

    for (ptr=ptr1; ptr<ptr2; ptr++) {
      sprintf(buf,"%02x",mem[ptr]);
      line += buf;
    }

    sprintf(buf,"0x%04x",ptr1);
    mxmlElementSetAttr(data,"addr",buf);
    mxmlNewText(data, 0, line.c_str());
}

void wxml(const char *filnam)
{
    FILE *fp;

    mxml_node_t *xml;    /* <?xml ... ?> */
    mxml_node_t *data;   /* <data> */
    mxml_node_t *group;  /* <group> */
    string tok;

    xml = mxmlNewXML("1.0");

    data = mxmlNewElement(xml, "?xml-stylesheet type=\"text/css\" href=\"sinclair.css\"?");

    data = mxmlNewElement(xml, "sinclair");
    mxmlElementSetAttr(data, "machine", "zx81");

    if (xml_binary) {

      group = mxmlNewElement(data, "hex");
      wxml_hex(group);

    } else {

      group = mxmlNewElement(data, "sysvars");
      wxml_sysvars(group);

      group = mxmlNewElement(data, "program");
      wxml_program(group);

      group = mxmlNewElement(data, "d_file");
      wxml_d_file(group);

      group = mxmlNewElement(data, "vars");
      wxml_vars(group);

    }

    fp = fopen(filnam, "w");
    mxmlSetWrapMargin(xml_wrap_margin);

    mxmlSaveFile(xml, fp, whitespace_cb);
    mxmlDelete(xml);

    fclose(fp);
}

mxml_type_t type_cb(mxml_node_t *node)
    {
    const char *name;
    mxml_type_t type;

    type = MXML_ELEMENT;
    name = mxmlGetElement(node);

    if (!strcmp(name, "statement") || !strcmp(name, "hex") || !strcmp(name,"sysvar") || !strcmp(name,"var") || !strcmp(name,"token"))
	type = MXML_TEXT;
    else if (!strcmp(name, "b")) type = MXML_OPAQUE;
    return type;
 }

void rxml_sinclair(mxml_node_t *node)
{
  mxml_node_t *snode;
  const char *attr, *name, *addr;
  int ptr;
  zline zl;
  string line;
  int istok, ntok;
  int ishex;
  char buf[4];
  int value;
  int i;

  if (!node) {
    fprintf(stderr,"zce: rxml_sinclair: no `<sinclair>` node found\n");
    return;
  }
  attr = mxmlElementGetAttr(node,"machine");
  if (!attr) {
    fprintf(stderr,"zce: rxml_sinclair: no `machine` attribute found\n");
    return;
  }
  if (strcmp(attr,"zx81")) {
    fprintf(stderr,"zce: rxml_sinclair: expected machine `zx81`\n");
    return;
  }
  line = "";
  snode = node;
  while (1) {
    snode = mxmlWalkNext(snode, node, MXML_DESCEND);
    if (!snode) break;
    name = mxmlGetElement(snode);
    if (name) {
      istok = ishex = 0;
      if (!strcmp(name,"b")) {
        if (line!="") { zl.set(line.c_str()); zl.compile(); zl.insert(); line=""; }
      } else if (!strcmp(name,"token")) {
	  istok = 1;
      } else if (!strcmp(name,"hex")) {
	ishex = 1;
        addr = mxmlElementGetAttr(snode,"addr");
      }
      else break;
    } else if (snode->type==MXML_TEXT) {
        if (istok) {
          attr = mxmlGetText(snode,0);
          ntok = str2tok(attr);
	  sprintf(buf,"#%02x",ntok);
          line += buf;
        } else if (ishex) {
          attr = mxmlGetText(snode,0);
	  if (addr) {
	    atoih(addr,ptr);
            for (i=0; i<(int)strlen(attr)/2; i++) {
	      htoi(attr+i*2,value);
	      mem[ptr+i] = (unsigned char) value;
	    }
	  } else {
            for (i=0; i<(int)strlen(attr)/2; i++) {
	      line += '#';
	      line += attr[i*2];
	      line += attr[i*2+1];
	    }
          }
	}
    } else if (snode->type==MXML_OPAQUE) {
        line += mxmlGetOpaque(snode);
    }
  }
  if (line!="") { zl.set(line.c_str()); zl.compile(); zl.insert(); }
}

/* if an addr and/or name in sysvar is given, the address could be checked and/or set
   according to zx conventions */

void rxml_sysvars(mxml_node_t *node)
{
  mxml_node_t *snode, *hnode;
  const char *name, *attr, *hex;
  int i, n, addr, bytes, value, hbytes, hvalue, hhex;

  snode = node;
  while (1) {
    snode = mxmlWalkNext(snode, node, MXML_DESCEND_FIRST);
    hnode = NULL;
    if (!snode) break;
    attr = mxmlGetElement(snode);
    if (strcmp(attr,"sysvar")) {
      fprintf(stderr,"zce: rxml_sysvars: unexpected `%s`\n",attr);
      return;
    } else {
      hbytes = hvalue = hhex = 0;
      name = mxmlElementGetAttr(snode,"name");
      attr = mxmlElementGetAttr(snode,"value");
      if (attr) {
        n = atoih(attr,value);
        if (!n)
          fprintf(stderr,"zce: rxml_sysvars: value `%s` could not be parsed\n",attr);
        else
	  hvalue = 1;
      } else {
        hnode = snode;
        hnode = mxmlWalkNext(hnode, node, MXML_DESCEND_FIRST);
        if (!hnode) break;
        hex = mxmlGetText(hnode,0);
        if (hex)
	  hhex = 1;
        else
	  fprintf(stderr,"zce: rxml_sysvars: value for `%s` needed\n",name);
      }
      attr = mxmlElementGetAttr(snode,"bytes");
      if (attr) {
        n = atoih(attr,bytes);
        if (!n)
          fprintf(stderr,"zce: rxml_sysvars: bytes `%s` could not be parsed\n",attr);
	else
          hbytes = 1;
      } else {
        fprintf(stderr,"zce: rxml_sysvars: bytes for `%s` needed\n",name);
      }
      attr = mxmlElementGetAttr(snode,"addr");
      if (attr) {
        n = atoih(attr,addr);
        if (!n)
          fprintf(stderr,"zce: rxml_sysvars: addr `%s` could not be parsed\n",attr);
      } else {
        fprintf(stderr,"zce: rxml_sysvars: no address for `%s`\n",name);
      }
      if (addr && hbytes && (hvalue || hhex)) {
        if ((unsigned int)addr != rptr) fprintf(stderr,"zce: rxml_sysvars: using address 0x%04x instead of 0x%04x\n",addr,rptr);
        if (hvalue) {
          switch (bytes) {
            case 1: mem[addr] = value; break;
            case 2: mem[addr] = value & 0x00ff; mem[addr+1] = (value & 0xff00) >> 8; break;
            default: fprintf(stderr,"zce: rxml_sysvars: invalid bytes `%d` for `%s`\n",bytes,name);
	  }
	} else {
          for (i=0; i<bytes; i++) {
            htoi(hex+i*2,value);
	    if (value>=0) mem[addr+i] = value;
	  }
        }
      }
      if (hbytes) rptr += bytes;
    }
    if (hnode) snode = hnode->child;
  }
}

void rxml_program(mxml_node_t *node)
{
  mxml_node_t *snode, *tnode;
  const char *attr, *txt;

  snode = node;
  while (1) {
    snode = mxmlWalkNext(snode, node, MXML_DESCEND_FIRST);
    tnode = NULL;
    if (!snode) break;
    attr = mxmlGetElement(snode);
    if (snode->type == MXML_TEXT) {
      txt = mxmlGetText(snode,0);
      printf("text %s\n",txt);
    } else if (!strcmp(attr,"statement")) {
      if (attr) printf("%s\n",attr);
      attr = mxmlElementGetAttr(snode,"addr");
      if (attr) printf("%s\n",attr);
    } else if (!strcmp(attr,"token")) {
      tnode = snode;
      tnode = mxmlWalkNext(tnode, node, MXML_DESCEND_FIRST);
      if (!tnode) break;
      txt = mxmlGetText(tnode,0);
      printf("token %s\n",txt);
    } else if (!strcmp(attr,"hex")) {
      tnode = snode;
      tnode = mxmlWalkNext(tnode, node, MXML_DESCEND_FIRST);
      if (!tnode) break;
      txt = mxmlGetText(tnode,0);
      printf("hex %s\n",txt);
    }
    if (tnode) snode = tnode;
  }
}

void rxml_d_file(mxml_node_t *node)
{
  mxml_node_t *snode;
  const char *attr;

  snode = node;
  while (1) {
    snode = mxmlWalkNext(snode, node, MXML_DESCEND_FIRST);
    if (!snode) break;
    attr = mxmlGetElement(snode);
    if (attr) printf("%s\n",attr);
    attr = mxmlElementGetAttr(snode,"addr");
    if (attr) printf("%s\n",attr);
    attr = mxmlElementGetAttr(snode,"bytes");
    if (attr) printf("%s\n",attr);
  }
}

void rxml_vars(mxml_node_t *node)
{
  mxml_node_t *snode;
  const char *attr;

  snode = node;
  while (1) {
    snode = mxmlWalkNext(snode, node, MXML_DESCEND_FIRST);
    if (!snode) break;
    attr = mxmlGetElement(snode);
    if (attr) printf("%s\n",attr);
    attr = mxmlElementGetAttr(snode,"addr");
    if (attr) printf("%s\n",attr);
    attr = mxmlElementGetAttr(snode,"bytes");
    if (attr) printf("%s\n",attr);
    attr = mxmlElementGetAttr(snode,"name");
    if (attr) printf("%s\n",attr);
    attr = mxmlElementGetAttr(snode,"value");
    if (attr) printf("%s\n",attr);
  }
}

void rxml(const char *filnam)
{
    FILE *fp;
    mxml_node_t *tree;
    mxml_node_t *node;

    rptr = 0x4009;
    fp = fopen(filnam, "r");
    tree = mxmlLoadFile(NULL, fp, type_cb);

    node = tree;
    node = mxmlFindElement(node,tree,"sinclair",NULL,NULL,MXML_DESCEND);
    rxml_sinclair(node);

    node = tree;
    node = mxmlFindElement(node,tree,"sysvars",NULL,NULL,MXML_DESCEND);
    rxml_sysvars(node);

    node = tree;
    node = mxmlFindElement(node,tree,"program",NULL,NULL,MXML_DESCEND);
    rxml_program(node);

    node = tree;
    node = mxmlFindElement(node,tree,"d_file",NULL,NULL,MXML_DESCEND);
    rxml_d_file(node);

    node = tree;
    node = mxmlFindElement(node,tree,"vars",NULL,NULL,MXML_DESCEND);
    rxml_vars(node);

    mxmlDelete(tree);
    fclose(fp);
}
